from __future__ import annotations

import datetime
import warnings
from typing import Any, Protocol

import toga
from toga.handlers import wrapped_handler

from .base import StyleT, Widget

# This accommodates the ranges of all existing implementations:
#  * datetime.date: 1 - 9999
#  * Android: approx 5,800,000 BC - 5,800,000
#  * Windows: 1753 - 9998
MIN_DATE = datetime.date(1800, 1, 1)
MAX_DATE = datetime.date(8999, 12, 31)


class OnChangeHandler(Protocol):
    def __call__(self, widget: DateInput, /, **kwargs: Any) -> object:
        """A handler that will be invoked when a change occurs.

        :param widget: The DateInput that was changed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class DateInput(Widget):
    _MIN_WIDTH = 200

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        value: datetime.date | None = None,
        min: datetime.date | None = None,
        max: datetime.date | None = None,
        on_change: toga.widgets.dateinput.OnChangeHandler | None = None,
    ):
        """Create a new DateInput widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param value: The initial date to display. If not specified, the current date
            will be used.
        :param min: The earliest date (inclusive) that can be selected.
        :param max: The latest date (inclusive) that can be selected.
        :param on_change: A handler that will be invoked when the value changes.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a DateInput
        self._impl = self.factory.DateInput(interface=self)

        self.on_change = None
        self.min = min
        self.max = max

        self.value = value
        self.on_change = on_change

    @property
    def value(self) -> datetime.date:
        """The currently selected date. A value of ``None`` will be converted into
        today's date.

        If this property is set to a value outside of the min/max range, it will be
        clipped.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value: object) -> None:
        value = self._convert_date(value, check_range=False)

        if value < self.min:
            value = self.min
        elif value > self.max:
            value = self.max

        self._impl.set_value(value)

    def _convert_date(self, value: object, *, check_range: bool) -> datetime.date:
        if value is None:
            value = datetime.date.today()
        elif isinstance(value, datetime.datetime):
            value = value.date()
        elif isinstance(value, datetime.date):
            pass
        elif isinstance(value, str):
            value = datetime.date.fromisoformat(value)
        else:
            raise TypeError("Not a valid date value")

        if check_range:
            if value < MIN_DATE:
                raise ValueError(f"The lowest supported date is {MIN_DATE.isoformat()}")
            if value > MAX_DATE:
                raise ValueError(
                    f"The highest supported date is {MAX_DATE.isoformat()}"
                )

        return value

    @property
    def min(self) -> datetime.date:
        """The minimum allowable date (inclusive). A value of ``None`` will be converted
        into the lowest supported date of 1800-01-01.

        When setting this property, the current :attr:`value` and :attr:`max` will be
        clipped against the new minimum value.

        :raises ValueError: If set to a date outside of the supported range.
        """
        return self._impl.get_min_date()

    @min.setter
    def min(self, value: object) -> None:
        if value is None:
            min = MIN_DATE
        else:
            min = self._convert_date(value, check_range=True)

        if self.max < min:
            self._impl.set_max_date(min)
        self._impl.set_min_date(min)
        if self.value < min:
            self.value = min

    @property
    def max(self) -> datetime.date:
        """The maximum allowable date (inclusive). A value of ``None`` will be converted
        into the highest supported date of 8999-12-31.

        When setting this property, the current :attr:`value` and :attr:`min` will be
        clipped against the new maximum value.

        :raises ValueError: If set to a date outside of the supported range.
        """
        return self._impl.get_max_date()

    @max.setter
    def max(self, value: object) -> None:
        if value is None:
            max = MAX_DATE
        else:
            max = self._convert_date(value, check_range=True)

        if self.min > max:
            self._impl.set_min_date(max)
        self._impl.set_max_date(max)
        if self.value > max:
            self.value = max

    @property
    def on_change(self) -> OnChangeHandler:
        """The handler to invoke when the date value changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler: toga.widgets.dateinput.OnChangeHandler) -> None:
        self._on_change = wrapped_handler(self, handler)


# 2023-05: Backwards compatibility
class DatePicker(DateInput):
    def __init__(self, *args: Any, **kwargs: Any):
        warnings.warn("DatePicker has been renamed DateInput.", DeprecationWarning)

        for old_name, new_name in [
            ("min_date", "min"),
            ("max_date", "max"),
        ]:
            try:
                value = kwargs.pop(old_name)
                warnings.warn(
                    f"DatePicker.{old_name} has been renamed DateInput.{new_name}",
                    DeprecationWarning,
                )
            except KeyError:
                pass
            else:
                kwargs[new_name] = value

        super().__init__(*args, **kwargs)

    @property
    def min_date(self) -> datetime.date:
        warnings.warn(
            "DatePicker.min_date has been renamed DateInput.min", DeprecationWarning
        )
        return self.min

    @min_date.setter
    def min_date(self, value: object) -> None:
        warnings.warn(
            "DatePicker.min_date has been renamed DateInput.min", DeprecationWarning
        )
        self.min = value

    @property
    def max_date(self) -> datetime.date:
        warnings.warn(
            "DatePicker.max_date has been renamed DateInput.max", DeprecationWarning
        )
        return self.max

    @max_date.setter
    def max_date(self, value: object) -> None:
        warnings.warn(
            "DatePicker.max_date has been renamed DateInput.max", DeprecationWarning
        )
        self.max = value



from __future__ import annotations

from .textinput import TextInput


class PasswordInput(TextInput):
    """Create a new password input widget."""

    def _create(self) -> None:
        self._impl = self.factory.PasswordInput(interface=self)



from __future__ import annotations

import importlib
import os
import sys
import warnings
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, TypeVar
from warnings import warn

import toga
from toga.platform import get_platform_factory

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    # Before Python 3.10, entry_points did not support the group argument;
    # so, the backport package must be used on older versions.
    from importlib_metadata import entry_points

# Make sure deprecation warnings are shown by default
warnings.filterwarnings("default", category=DeprecationWarning)

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias

    # Define a type variable for generics where an Image type is required.
    ImageT = TypeVar("ImageT")

    # Define the types that can be used as Image content
    PathLikeT: TypeAlias = str | os.PathLike
    BytesLikeT: TypeAlias = bytes | bytearray | memoryview
    ImageLikeT: TypeAlias = Any
    ImageContentT: TypeAlias = PathLikeT | BytesLikeT | ImageLikeT

    # Define a type variable representing an image of an externally defined type.
    ExternalImageT = TypeVar("ExternalImageT")


class ImageConverter(Protocol):
    """A class to convert between an externally defined image type and
    :any:`toga.Image`.
    """

    #: The base image class this plugin can interpret.
    image_class: type[ExternalImageT]

    @staticmethod
    def convert_from_format(image_in_format: ExternalImageT) -> BytesLikeT:
        """Convert from :any:`image_class` to data in a :ref:`known image format
        <known-image-formats>`.

        Will accept an instance of :any:`image_class`, or subclass of that class.

        :param image_in_format: An instance of :any:`image_class` (or a subclass).
        :returns: The image data, in a :ref:`known image format <known-image-formats>`.
        """

    @staticmethod
    def convert_to_format(
        data: BytesLikeT,
        image_class: type[ExternalImageT],
    ) -> ExternalImageT:
        """Convert from data to :any:`image_class` or specified subclass.

        Accepts a bytes-like object representing the image in a
        :ref:`known image format <known-image-formats>`, and returns an instance of the
        image class specified. This image class is guaranteed to be either the
        :any:`image_class` registered by the plugin, or a subclass of that class.

        :param data: Image data in a :ref:`known image format <known-image-formats>`.
        :param image_class: The class of image to return.
        :returns: The image, as an instance of the image class specified.
        """


NOT_PROVIDED = object()


class Image:
    def __init__(
        self,
        src: ImageContentT = NOT_PROVIDED,
        *,
        path: object = NOT_PROVIDED,  # DEPRECATED
        data: object = NOT_PROVIDED,  # DEPRECATED
    ):
        """Create a new image.

        :param src: The source from which to load the image. Can be any valid
            :any:`image content <ImageContentT>` type.
        :param path: **DEPRECATED** - Use ``src``.
        :param data: **DEPRECATED** - Use ``src``.
        :raises FileNotFoundError: If a path is provided, but that path does not exist.
        :raises ValueError: If the source cannot be loaded as an image.
        """
        ######################################################################
        # 2023-11: Backwards compatibility
        ######################################################################
        num_provided = sum(arg is not NOT_PROVIDED for arg in (src, path, data))
        if num_provided > 1:
            raise TypeError("Received multiple arguments to constructor.")
        if num_provided == 0:
            raise TypeError(
                "Image.__init__() missing 1 required positional argument: 'src'"
            )
        if path is not NOT_PROVIDED:
            src = path
            warn(
                "Path argument is deprecated, use src instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        elif data is not NOT_PROVIDED:
            src = data
            warn(
                "Data argument is deprecated, use src instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        ######################################################################
        # End backwards compatibility
        ######################################################################

        self.factory = get_platform_factory()
        self._path = None

        # Any "lump of bytes" should be valid here.
        if isinstance(src, (bytes, bytearray, memoryview)):
            self._impl = self.factory.Image(interface=self, data=src)

        elif isinstance(src, (str, Path)):
            self._path = toga.App.app.paths.app / src
            if not self._path.is_file():
                raise FileNotFoundError(f"Image file {self._path} does not exist")
            self._impl = self.factory.Image(interface=self, path=self._path)

        elif isinstance(src, Image):
            self._impl = self.factory.Image(interface=self, data=src.data)

        elif isinstance(src, self.factory.Image.RAW_TYPE):
            self._impl = self.factory.Image(interface=self, raw=src)

        else:
            for converter in self._converters():
                if isinstance(src, converter.image_class):
                    data = converter.convert_from_format(src)
                    self._impl = self.factory.Image(interface=self, data=data)
                    return

            raise TypeError("Unsupported source type for Image")

    @classmethod
    @lru_cache(maxsize=None)
    def _converters(cls) -> list[ImageConverter]:
        """Return list of registered image plugin converters. Only loaded once."""
        converters = []

        for image_plugin in entry_points(group="toga.image_formats"):
            module_name, class_name = image_plugin.value.rsplit(".", 1)
            module = importlib.import_module(module_name)
            converter = getattr(module, class_name)

            if converter.image_class is not None:
                converters.append(converter)

        return converters

    @property
    def size(self) -> tuple[int, int]:
        """The size of the image, as a (width, height) tuple."""
        return self._impl.get_width(), self._impl.get_height()

    @property
    def width(self) -> int:
        """The width of the image, in pixels."""
        return self._impl.get_width()

    @property
    def height(self) -> int:
        """The height of the image, in pixels."""
        return self._impl.get_height()

    @property
    def data(self) -> bytes:
        """The raw data for the image, in PNG format."""
        return self._impl.get_data()

    @property
    def path(self) -> Path | None:
        """The path from which the image was opened, if any (or None)."""
        return self._path

    def save(self, path: str | Path) -> None:
        """Save image to given path.

        The file format of the saved image will be determined by the extension of
        the filename provided (e.g ``path/to/mypicture.png`` will save a PNG file).

        :param path: Path to save the image to.
        """
        self._impl.save(path)

    def as_format(self, format: type[ImageT]) -> ImageT:
        """Return the image, converted to the image format specified.

        :param format: Format to provide. Defaults to :class:`~toga.images.Image`; also
             supports :any:`PIL.Image.Image` if Pillow is installed, as well as any image
             types defined by installed :doc:`image format plugins
             </reference/plugins/image_formats>`.
        :returns: The image in the requested format
        :raises TypeError: If the format supplied is not recognized.
        """
        if isinstance(format, type):
            if issubclass(format, Image):
                return format(self.data)

            for converter in self._converters():
                if issubclass(format, converter.image_class):
                    return converter.convert_to_format(self.data, format)

        raise TypeError(f"Unknown conversion format for Image: {format}")



import os

# Only set up the app processes for coverage *if* coverage is running.
# Otherwise, just importing coverage will force running coverage.
# The COVERAGE_RUN environment variable is set by coverage.py itself.
if os.environ.get("COVERAGE_RUN"):
    import coverage

    coverage.process_startup()



import pytest

import toga
from toga.colors import REBECCAPURPLE
from toga.fonts import FANTASY
from toga.style.pack import HIDDEN, RIGHT, VISIBLE
from toga_dummy.utils import assert_action_performed_with


# Create the simplest possible widget with a concrete implementation that will
# allow children
class ExampleWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)
        self._children = []


# Create the simplest possible widget with a concrete implementation that cannot
# have children.
class ExampleLeafWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)


@pytest.fixture
def grandchild():
    return ExampleLeafWidget(id="grandchild_id")


@pytest.fixture
def child(grandchild):
    child = ExampleWidget(id="child_id")
    child.add(grandchild)

    return child


@pytest.fixture
def widget(child):
    widget = ExampleWidget(id="widget_id")
    widget.add(child)

    return widget


def test_refresh(widget):
    """Refresh requests are passed to the widget."""
    widget.applicator.refresh()

    assert_action_performed_with(widget, "refresh")


def test_set_bounds(widget, child, grandchild):
    """Bounds changes are passed to all widgets in the tree."""
    # Manually set location of the parent
    widget.layout._origin_left = 100
    widget.layout._origin_top = 200
    widget.layout.content_width = 300
    widget.layout.content_height = 400

    # Manually set location of the child
    child.layout._origin_left = 10
    child.layout._origin_top = 20
    child.layout.content_width = 30
    child.layout.content_height = 40

    # Manually set location of the grandchild
    grandchild.layout._origin_left = 1
    grandchild.layout._origin_top = 2
    grandchild.layout.content_width = 3
    grandchild.layout.content_height = 4

    # Propagate the bounds
    widget.applicator.set_bounds()

    assert_action_performed_with(
        widget, "set bounds", x=100, y=200, width=300, height=400
    )
    assert_action_performed_with(child, "set bounds", x=10, y=20, width=30, height=40)
    assert_action_performed_with(grandchild, "set bounds", x=1, y=2, width=3, height=4)


def test_text_alignment(widget):
    """Text alignment can be set on a widget."""
    widget.applicator.set_text_alignment(RIGHT)

    assert_action_performed_with(widget, "set alignment", alignment=RIGHT)


@pytest.mark.parametrize(
    "child_visibility, grandchild_visibility, value, "
    "widget_hidden, child_hidden, grandchild_hidden",
    [
        # Set widget hidden. All widgets are always hidden,
        # no matter the actual style setting.
        (VISIBLE, VISIBLE, True, True, True, True),
        (VISIBLE, HIDDEN, True, True, True, True),
        (HIDDEN, VISIBLE, True, True, True, True),
        (HIDDEN, HIDDEN, True, True, True, True),
        # Set widget visible. Visibility only cascades
        # as far as the first HIDDEN widget
        (VISIBLE, VISIBLE, False, False, False, False),
        (VISIBLE, HIDDEN, False, False, False, True),
        (HIDDEN, VISIBLE, False, False, True, True),
        (HIDDEN, HIDDEN, False, False, True, True),
    ],
)
def test_set_hidden(
    widget,
    child,
    grandchild,
    child_visibility,
    grandchild_visibility,
    value,
    widget_hidden,
    child_hidden,
    grandchild_hidden,
):
    """Widget visibility can be controlled, and is transitive into children."""
    # Set the explicit visibility of the child and grandchild
    child.style.visibility = child_visibility
    grandchild.style.visibility = grandchild_visibility

    # Set widget visibility
    widget.applicator.set_hidden(value)

    assert_action_performed_with(widget, "set hidden", hidden=widget_hidden)
    assert_action_performed_with(child, "set hidden", hidden=child_hidden)
    assert_action_performed_with(grandchild, "set hidden", hidden=grandchild_hidden)

    # The style property of the child and grandchild hasn't changed.
    assert child.style.visibility == child_visibility
    assert grandchild.style.visibility == grandchild_visibility


def test_set_font(widget):
    """A font change can be applied to a widget."""
    widget.applicator.set_font(FANTASY)

    assert_action_performed_with(widget, "set font", font=FANTASY)


def test_set_color(widget):
    """A color change can be applied to a widget."""
    widget.applicator.set_color(REBECCAPURPLE)

    assert_action_performed_with(widget, "set color", color=REBECCAPURPLE)


def test_set_background_color(child, widget):
    """A background color change can be applied to a widget."""
    widget.applicator.set_background_color(REBECCAPURPLE)

    assert_action_performed_with(widget, "set background color", color=REBECCAPURPLE)



from __future__ import annotations

import re
from abc import abstractmethod


class BooleanValidator:
    def __init__(self, error_message: str, allow_empty: bool = True):
        """An abstract base class for defining a simple validator.

        Subclasses should implement the ``is_valid()`` method

        :param error_message: The error to display to the user when the input
            isn't valid.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        self.error_message = error_message
        self.allow_empty = allow_empty

    def __call__(self, input_string: str) -> str | None:
        if self.allow_empty and input_string == "":
            return None
        return None if self.is_valid(input_string) else self.error_message

    @abstractmethod
    def is_valid(self, input_string: str) -> bool:
        """Is the input string valid?

        :param input_string: The string to validate.
        :returns: ``True`` if the input is valid.
        """


class CountValidator:
    def __init__(
        self,
        count: int | None,
        expected_existence_message: str,
        expected_non_existence_message: str,
        expected_count_message: str,
        allow_empty: bool = True,
    ):
        """An abstract base class for validators that are based on counting
        instances of some content in the overall content.

        Subclasses should implement the ``count()`` method to identify the
        content of interest.

        :param count: Optional; The expected count.
        :param expected_existence_message: The error message to show if matches are
            expected, but were not found.
        :param expected_non_existence_message: The error message to show if matches were
            not expected, but were found.
        :param expected_count_message: The error message to show if matches
            were expected, but a different number were found.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        self.expected_count = count
        self.expected_existence_message = expected_existence_message
        self.expected_non_existence_message = expected_non_existence_message
        self.expected_count_message = expected_count_message
        self.allow_empty = allow_empty

    def __call__(self, input_string: str) -> str | None:
        if self.allow_empty and input_string == "":
            return None
        actual_count = self.count(input_string)
        if actual_count == 0 and self.expected_count != 0:
            return self.expected_existence_message
        if actual_count != 0 and self.expected_count == 0:
            return self.expected_non_existence_message
        if self.expected_count is not None and actual_count != self.expected_count:
            return self.expected_count_message
        return None

    @abstractmethod
    def count(self, input_string: str) -> int:
        """Count the instances of content of interest in the input string.

        :param input_string: The string to inspect for content of interest.
        :returns: The number of instances of content that the validator is looking for.
        """


class LengthBetween(BooleanValidator):
    def __init__(
        self,
        min_length: int | None,
        max_length: int | None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the length of input falls in a given
        range.

        :param min_length: The minimum length of the string (inclusive).
        :param max_length: The maximum length of the string (inclusive).
        :param error_message: Optional; the error message to display when the
            length isn't in the given range.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = (
                f"Input should be between {min_length} and {max_length} characters"
            )
        super().__init__(error_message=error_message, allow_empty=allow_empty)

        if (
            min_length is not None
            and max_length is not None
            and min_length > max_length
        ):
            raise ValueError("Minimum length cannot be less than maximum length")

        self.min_length = min_length
        self.max_length = max_length

    def is_valid(self, input_string: str) -> bool:
        if self.min_length is not None:
            if len(input_string) < self.min_length:
                return False
        if self.max_length is not None:
            if len(input_string) > self.max_length:
                return False
        return True


class MinLength(LengthBetween):
    def __init__(
        self,
        length: int,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the length of input exceeds a minimum size.

        :param length: The minimum length of the string (inclusive).
        :param error_message: Optional; the error message to display when the
            string isn't long enough.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = f"Input is too short (length should be at least {length})"
        super().__init__(
            min_length=length,
            max_length=None,
            error_message=error_message,
            allow_empty=allow_empty,
        )


class MaxLength(LengthBetween):
    def __init__(
        self,
        length: int,
        error_message: str | None = None,
    ):
        """A validator confirming that the length of input does not exceed a maximum size.

        :param length: The maximum length of the string (inclusive).
        :param error_message: Optional; the error message to display when the
            string is too long.
        """
        if error_message is None:
            error_message = f"Input is too long (length should be at most {length})"
        super().__init__(
            min_length=None,
            max_length=length,
            error_message=error_message,
        )


class StartsWith(BooleanValidator):
    def __init__(
        self,
        substring: str,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the input starts with a given substring.

        :param substring: The substring that the input must start with.
        :param error_message: Optional; the error message to display when the
            string doesn't start with the given substring.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = f"Input should start with {substring!r}"

        super().__init__(error_message=error_message, allow_empty=allow_empty)
        self.substring = substring

    def is_valid(self, input_string: str) -> bool:
        return input_string.startswith(self.substring)


class EndsWith(BooleanValidator):
    def __init__(
        self,
        substring: str,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string ends with a given substring.

        :param substring: The substring that the input must end with.
        :param error_message: Optional; the error message to display when the
            string doesn't end with the given substring.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = f"Input should end with '{substring}'"

        super().__init__(error_message=error_message, allow_empty=allow_empty)
        self.substring = substring

    def is_valid(self, input_string: str) -> bool:
        return input_string.endswith(self.substring)


class Contains(CountValidator):
    def __init__(
        self,
        substring: str,
        count: int | None = None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string contains one or more
        copies of a substring.

        :param substring: The substring that must exist in the input.
        :param count: Optional; The exact number of matches that are expected.
        :param error_message: Optional; the error message to display when the
            input doesn't contain the substring (or the requested count of substrings).
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is not None:
            expected_non_existence_message = error_message
            expected_count_message = error_message
            expected_existence_message = error_message
        else:
            expected_existence_message = f"Input should contain {substring!r}"
            expected_non_existence_message = f"Input should not contain {substring!r}"
            expected_count_message = (
                f"Input should contain {substring!r} exactly {count} times"
            )

        super().__init__(
            count=count,
            expected_existence_message=expected_existence_message,
            expected_non_existence_message=expected_non_existence_message,
            expected_count_message=expected_count_message,
            allow_empty=allow_empty,
        )
        self.substring = substring

    def count(self, input_string: str) -> int:
        return input_string.count(self.substring)


class NotContains(Contains):
    def __init__(
        self,
        substring: str,
        error_message: str | None = None,
    ):
        """A validator confirming that the string does not contain a substring.

        :param substring: A substring that must not exist in the input.
        :param error_message: Optional; the error message to display when the
            input contains the provided substring.
        """
        super().__init__(
            substring,
            count=0,
            error_message=error_message,
        )


class MatchRegex(BooleanValidator):
    def __init__(
        self,
        regex_string: str,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string matches a given regular expression.

        :param regex_string: A regular expression that the input must match.
        :param error_message: Optional; the error message to display when the
            input doesn't match the provided regular expression.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = f"Input should match regex: {regex_string!r}"

        super().__init__(error_message=error_message, allow_empty=allow_empty)
        self.regex_string = regex_string

    def is_valid(self, input_string: str) -> bool:
        return bool(re.search(self.regex_string, input_string))


class ContainsUppercase(CountValidator):
    def __init__(
        self,
        count: int | None = None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string contains upper case letters.

        :param count: Optional; if provided, the exact count of upper
            case letters that must exist. If not provided, the existence of any
            upper case letter will make the string valid.
        :param error_message: Optional; the error message to display when the
            input doesn't contain upper case letters (or the requested count of
            upper case letters).
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is not None:
            expected_non_existence_message = error_message
            expected_count_message = error_message
            expected_existence_message = error_message
        else:
            expected_existence_message = (
                "Input should contain at least one upper case character"
            )
            expected_non_existence_message = (
                "Input should not contain upper case characters"
            )
            expected_count_message = (
                f"Input should contain exactly {count} upper case characters"
            )

        super().__init__(
            count=count,
            expected_existence_message=expected_existence_message,
            expected_non_existence_message=expected_non_existence_message,
            expected_count_message=expected_count_message,
            allow_empty=allow_empty,
        )

    def count(self, input_string: str) -> int:
        return len([char for char in input_string if char.isupper()])


class ContainsLowercase(CountValidator):
    def __init__(
        self,
        count: int | None = None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string contains lower case letters.

        :param count: Optional; if provided, the exact count of lower
            case letters that must exist. If not provided, the existence of any
            lower case letter will make the string valid.
        :param error_message: Optional; the error message to display when the
            input doesn't contain lower case letters (or the requested count of
            lower case letters).
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is not None:
            expected_non_existence_message = error_message
            expected_count_message = error_message
            expected_existence_message = error_message
        else:
            expected_existence_message = (
                "Input should contain at least one lower case character"
            )
            expected_non_existence_message = (
                "Input should not contain lower case characters"
            )
            expected_count_message = (
                f"Input should contain exactly {count} lower case characters"
            )

        super().__init__(
            count=count,
            expected_existence_message=expected_existence_message,
            expected_non_existence_message=expected_non_existence_message,
            expected_count_message=expected_count_message,
            allow_empty=allow_empty,
        )

    def count(self, input_string: str) -> int:
        return len([char for char in input_string if char.islower()])


class ContainsDigit(CountValidator):
    def __init__(
        self,
        count: int | None = None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string contains digits.

        :param count: Optional; if provided, the exact count of digits
            that must exist. If not provided, the existence of any digit will
            make the string valid.
        :param error_message: Optional; the error message to display when the
            input doesn't contain digits (or the requested count of digits).
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is not None:
            expected_non_existence_message = error_message
            expected_count_message = error_message
            expected_existence_message = error_message
        else:
            expected_existence_message = "Input should contain at least one digit"
            expected_non_existence_message = "Input should not contain digits"
            expected_count_message = f"Input should contain exactly {count} digits"

        super().__init__(
            count=count,
            expected_existence_message=expected_existence_message,
            expected_non_existence_message=expected_non_existence_message,
            expected_count_message=expected_count_message,
            allow_empty=allow_empty,
        )

    def count(self, input_string: str) -> int:
        return len([char for char in input_string if char.isdigit()])


class ContainsSpecial(CountValidator):
    def __init__(
        self,
        count: int | None = None,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string contains "special" (i.e.,
        non-alphanumeric) characters.

        :param count: Optional; if provided, the exact count of special
            characters that must exist. If not provided, the existence of any
            special character will make the string valid.
        :param error_message: Optional; the error message to display when the
            input doesn't contain special characters (or the requested count of
            special characters).
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is not None:
            expected_non_existence_message = error_message
            expected_count_message = error_message
            expected_existence_message = error_message
        else:
            expected_existence_message = (
                "Input should contain at least one special character"
            )
            expected_non_existence_message = (
                "Input should not contain special characters"
            )
            expected_count_message = (
                f"Input should contain exactly {count} special characters"
            )

        super().__init__(
            count=count,
            expected_existence_message=expected_existence_message,
            expected_non_existence_message=expected_non_existence_message,
            expected_count_message=expected_count_message,
            allow_empty=allow_empty,
        )

    def count(self, input_string: str) -> int:
        return len(
            [
                char
                for char in input_string
                if not char.isalpha() and not char.isdigit() and not char.isspace()
            ]
        )


class Integer(BooleanValidator):
    def __init__(
        self,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string is an integer.

        :param error_message: Optional; the error message to display when the
            input isn't an integer.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = "Input should be an integer"
        super().__init__(error_message=error_message, allow_empty=allow_empty)

    def is_valid(self, input_string: str) -> bool:
        try:
            int(input_string)
            return True
        except ValueError:
            return False


class Number(BooleanValidator):
    def __init__(
        self,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string is a number.

        :param error_message: Optional; the error message to display when the
            input isn't a number.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = "Input should be a number"
        super().__init__(error_message=error_message, allow_empty=allow_empty)

    def is_valid(self, input_string: str) -> bool:
        try:
            float(input_string)
            return True
        except ValueError:
            return False


class Email(MatchRegex):
    EMAIL_REGEX = (
        r"^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$"
    )

    def __init__(
        self,
        error_message: str | None = None,
        allow_empty: bool = True,
    ):
        """A validator confirming that the string is an email address.

        .. note::

            It's impossible to do *true* RFC-compliant email validation with a regex.
            This validator does a "best effort" validation. It will inevitably allow
            some email addresses that aren't *technically* valid. However, it shouldn't
            *exclude* any valid email addresses.

        :param error_message: Optional; the error message to display when the
            input isn't a number.
        :param allow_empty: Optional; Is no input considered valid? Defaults to
            ``True``
        """
        if error_message is None:
            error_message = "Input should be a valid email address"
        super().__init__(
            self.EMAIL_REGEX, error_message=error_message, allow_empty=allow_empty
        )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from toga.app import App
from toga.constants import Direction
from toga.window import Window

from .base import StyleT, Widget

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import TypeAlias
    else:
        from typing import TypeAlias

    SplitContainerContentT: TypeAlias = Widget | tuple[Widget, float] | None


class SplitContainer(Widget):
    HORIZONTAL = Direction.HORIZONTAL
    VERTICAL = Direction.VERTICAL

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        direction: Direction = Direction.VERTICAL,
        content: tuple[SplitContainerContentT, SplitContainerContentT] = (None, None),
    ):
        """Create a new SplitContainer.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param direction: The direction in which the divider will be drawn. Either
            :attr:`~toga.constants.Direction.HORIZONTAL` or
            :attr:`~toga.constants.Direction.VERTICAL`; defaults to
            :attr:`~toga.constants.Direction.VERTICAL`
        :param content: Initial :any:`SplitContainer content <SplitContainerContentT>`
            of the container. Defaults to both panels being empty.
        """
        super().__init__(id=id, style=style)
        self._content: tuple[SplitContainerContentT, SplitContainerContentT] = (
            None,
            None,
        )

        # Create a platform specific implementation of a SplitContainer
        self._impl = self.factory.SplitContainer(interface=self)

        self.content = content
        self.direction = direction

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        SplitContainer widgets cannot be disabled; this property will always return
        True; any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    def focus(self) -> None:
        """No-op; SplitContainer cannot accept input focus."""
        pass

    @property
    def content(self) -> tuple[SplitContainerContentT, SplitContainerContentT]:
        """The widgets displayed in the SplitContainer.

        This property accepts a sequence of exactly 2 elements, each of which can be
        either:

        * A :any:`Widget` to display in the panel.
        * ``None``, to make the panel empty.
        * A tuple consisting of a Widget (or ``None``) and the initial flex value to
          apply to that panel in the split, which must be greater than 0.

        If a flex value isn't specified, a value of 1 is assumed.

        When reading this property, only the widgets are returned, not the flex values.
        """
        return self._content

    @content.setter
    def content(
        self, content: tuple[SplitContainerContentT, SplitContainerContentT]
    ) -> None:
        try:
            if len(content) != 2:
                raise TypeError()
        except TypeError:
            raise ValueError(
                "SplitContainer content must be a sequence with exactly 2 elements"
            )

        _content = []
        flex = []
        for item in content:
            if isinstance(item, tuple):
                if len(item) == 2:
                    widget, flex_value = item
                    if flex_value <= 0:
                        raise ValueError(
                            "The flex value for an item in a SplitContainer must be >0"
                        )
                else:
                    raise ValueError(
                        "An item in SplitContainer content must be a 2-tuple "
                        "containing the widget, and the flex weight to assign to that "
                        "widget."
                    )
            else:
                widget = item
                flex_value = 1

            _content.append(widget)
            flex.append(flex_value)

            if widget:
                widget.app = self.app
                widget.window = self.window

        self._impl.set_content(
            tuple(w._impl if w is not None else None for w in _content),
            flex,
        )
        self._content = tuple(_content)
        self.refresh()

    @Widget.app.setter
    def app(self, app: App | None) -> None:
        # Invoke the superclass property setter
        Widget.app.fset(self, app)

        # Also assign the app to the content in the container
        for content in self.content:
            if content:
                content.app = app

    @Widget.window.setter
    def window(self, window: Window | None) -> None:
        # Invoke the superclass property setter
        Widget.window.fset(self, window)

        # Also assign the window to the content in the container
        for content in self._content:
            if content:
                content.window = window

    @property
    def direction(self) -> Direction:
        """The direction of the split."""
        return self._impl.get_direction()

    @direction.setter
    def direction(self, value: object) -> None:
        self._impl.set_direction(value)
        self.refresh()



import pytest

from toga.colors import REBECCAPURPLE, rgb
from toga.constants import FillRule
from toga.widgets.canvas import (
    ClosedPathContext,
    Context,
    FillContext,
    LineTo,
    StrokeContext,
)
from toga_dummy.utils import assert_action_performed

REBECCA_PURPLE_COLOR = rgb(102, 51, 153)


def test_subcontext(widget):
    """A context can produce a subcontext."""
    with widget.context.Context() as subcontext:
        subcontext.line_to(30, 40)
    # A fresh context has been created as a subcontext of the canvas.
    assert isinstance(subcontext, Context)
    assert subcontext != widget.context

    assert_action_performed(widget, "redraw")
    assert repr(subcontext) == "Context()"

    # The first and last instructions can be ignored; they're the root canvas context
    assert widget._impl.draw_instructions[1:-1] == [
        ("push context", {}),
        ("line to", {"x": 30, "y": 40}),
        ("pop context", {}),
    ]


@pytest.mark.parametrize(
    "kwargs, args_repr, has_move, properties",
    [
        # Defaults
        (
            {},
            "x=None, y=None",
            False,
            {"x": None, "y": None},
        ),
        # X only
        (
            {"x": 10},
            "x=10, y=None",
            False,
            {"x": 10, "y": None},
        ),
        # Y only
        (
            {"y": 20},
            "x=None, y=20",
            False,
            {"x": None, "y": 20},
        ),
        # X and Y
        (
            {"x": 10, "y": 20},
            "x=10, y=20",
            True,
            {"x": 10, "y": 20},
        ),
    ],
)
def test_closed_path(widget, kwargs, args_repr, has_move, properties):
    """A context can produce a ClosedPath subcontext."""
    with widget.context.ClosedPath(**kwargs) as closed_path:
        closed_path.line_to(30, 40)

    # A fresh context has been created as a subcontext of the canvas.
    assert isinstance(closed_path, ClosedPathContext)
    assert repr(closed_path) == f"ClosedPathContext({args_repr})"

    assert_action_performed(widget, "redraw")

    # All the attributes can be retrieved.
    for attr, value in properties.items():
        assert getattr(closed_path, attr) == value

    # The first and last instructions can be ignored; they're the root canvas context
    assert widget._impl.draw_instructions[1:-1] == [
        ("push context", {}),
        ("begin path", {}),
    ] + ([("move to", {"x": 10, "y": 20})] if has_move else []) + [
        ("line to", {"x": 30, "y": 40}),
        ("close path", {}),
        ("pop context", {}),
    ]


@pytest.mark.parametrize(
    "kwargs, args_repr, has_move, properties",
    [
        # Defaults
        (
            {},
            "x=None, y=None, color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        # X only
        (
            {"x": 10},
            "x=10, y=None, color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            False,
            {"x": 10, "y": None, "color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # Y only
        (
            {"y": 20},
            "x=None, y=20, color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            False,
            {"x": None, "y": 20, "color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # X and Y
        (
            {"x": 10, "y": 20},
            "x=10, y=20, color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            True,
            {"x": 10, "y": 20, "color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # Color
        (
            {"color": REBECCAPURPLE},
            f"x=None, y=None, color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.NONZERO",
            False,
            {
                "x": None,
                "y": None,
                "color": REBECCA_PURPLE_COLOR,
                "fill_rule": FillRule.NONZERO,
            },
        ),
        # Reset color with None
        (
            {"color": None},
            "x=None, y=None, color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        # Fill Rule
        (
            {"x": None, "y": None, "fill_rule": FillRule.EVENODD},
            "x=None, y=None, color=rgb(0, 0, 0), fill_rule=FillRule.EVENODD",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "fill_rule": FillRule.EVENODD,
            },
        ),
        # All args
        (
            {"x": 10, "y": 20, "color": REBECCAPURPLE, "fill_rule": FillRule.EVENODD},
            f"x=10, y=20, color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.EVENODD",
            True,
            {
                "x": 10,
                "y": 20,
                "color": REBECCA_PURPLE_COLOR,
                "fill_rule": FillRule.EVENODD,
            },
        ),
    ],
)
def test_fill(widget, kwargs, args_repr, has_move, properties):
    """A context can produce a Fill subcontext."""
    with widget.context.Fill(**kwargs) as fill:
        fill.line_to(30, 40)

    # A fresh context has been created as a subcontext of the canvas.
    assert isinstance(fill, FillContext)
    assert repr(fill) == f"FillContext({args_repr})"

    # All the attributes can be retrieved.
    for attr, value in properties.items():
        assert getattr(fill, attr) == value

    # x and y aren't part of the fill instruction
    x = properties.pop("x")
    y = properties.pop("y")

    # The first and last instructions can be ignored; they're the root canvas context
    assert widget._impl.draw_instructions[1:-1] == [
        ("push context", {}),
        ("begin path", {}),
    ] + ([("move to", {"x": x, "y": y})] if has_move else []) + [
        (
            "line to",
            {
                "x": 30,
                "y": 40,
                "fill_color": properties["color"],
                "fill_rule": properties["fill_rule"],
            },
        ),
        ("fill", properties),
        ("pop context", {}),
    ]


@pytest.mark.parametrize(
    "kwargs, args_repr, has_move, properties",
    [
        # Defaults
        (
            {},
            "x=None, y=None, color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # X only
        (
            {"x": 10},
            "x=10, y=None, color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            False,
            {
                "x": 10,
                "y": None,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # Y only
        (
            {"y": 20},
            "x=None, y=20, color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            False,
            {
                "x": None,
                "y": 20,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # X and Y
        (
            {"x": 10, "y": 20},
            "x=10, y=20, color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            True,
            {
                "x": 10,
                "y": 20,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # Color
        (
            {"color": REBECCAPURPLE},
            f"x=None, y=None, color={REBECCA_PURPLE_COLOR}, line_width=2.0, line_dash=None",
            False,
            {
                "x": None,
                "y": None,
                "color": REBECCA_PURPLE_COLOR,
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # Reset color with None
        (
            {"color": None},
            "x=None, y=None, color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # Line width
        (
            {"x": None, "y": None, "line_width": 4.5},
            "x=None, y=None, color=rgb(0, 0, 0), line_width=4.5, line_dash=None",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "line_width": 4.5,
                "line_dash": None,
            },
        ),
        # Line dash
        (
            {"x": None, "y": None, "line_dash": [2, 7]},
            "x=None, y=None, color=rgb(0, 0, 0), line_width=2.0, line_dash=[2, 7]",
            False,
            {
                "x": None,
                "y": None,
                "color": rgb(0, 0, 0),
                "line_width": 2.0,
                "line_dash": [2, 7],
            },
        ),
        # All args
        (
            {
                "x": 10,
                "y": 20,
                "color": REBECCAPURPLE,
                "line_width": 4.5,
                "line_dash": [2, 7],
            },
            f"x=10, y=20, color={REBECCA_PURPLE_COLOR}, line_width=4.5, line_dash=[2, 7]",
            True,
            {
                "x": 10,
                "y": 20,
                "color": REBECCA_PURPLE_COLOR,
                "line_width": 4.5,
                "line_dash": [2, 7],
            },
        ),
    ],
)
def test_stroke(widget, kwargs, args_repr, has_move, properties):
    """A context can produce a Stroke subcontext."""
    with widget.context.Stroke(**kwargs) as stroke:
        stroke.line_to(30, 40)

    # A fresh context has been created as a subcontext of the canvas.
    assert isinstance(stroke, StrokeContext)
    assert repr(stroke) == f"StrokeContext({args_repr})"

    # All the attributes can be retrieved.
    for attr, value in properties.items():
        assert getattr(stroke, attr) == value

    # x and y aren't part of the stroke instruction
    x = properties.pop("x")
    y = properties.pop("y")

    # The first and last instructions can be ignored; they're the root canvas context
    assert widget._impl.draw_instructions[1:-1] == [
        ("push context", {}),
        ("begin path", {}),
    ] + ([("move to", {"x": x, "y": y})] if has_move else []) + [
        (
            "line to",
            {
                "x": 30,
                "y": 40,
                "stroke_color": properties["color"],
                "line_width": properties["line_width"],
                "line_dash": properties["line_dash"],
            },
        ),
        ("stroke", properties),
        ("pop context", {}),
    ]


def test_deprecated_drawing_operations(widget):
    """Deprecated simple drawing operations raise a warning."""

    with pytest.warns(
        DeprecationWarning,
        match=r"Context.new_path\(\) has been renamed Context.begin_path\(\)",
    ):
        widget.context.new_path()

    with pytest.warns(
        DeprecationWarning,
        match=r"Context.context\(\) has been renamed Context.Context\(\)",
    ):
        with widget.context.context() as subcontext:
            subcontext.line_to(20, 30)
            subcontext.line_to(30, 20)

    with pytest.warns(
        DeprecationWarning,
        match=r"Context.closed_path\(\) has been renamed Context.ClosedPath\(\)",
    ):
        with widget.context.closed_path(10, 20) as closed_path:
            closed_path.line_to(20, 30)
            closed_path.line_to(30, 20)

    # Assert the deprecated draw instructions rendered as expected
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("begin path", {}),
        # Context
        ("push context", {}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 30, "y": 20}),
        ("pop context", {}),
        # ClosedPath
        ("push context", {}),
        ("begin path", {}),
        ("move to", {"x": 10, "y": 20}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 30, "y": 20}),
        ("close path", {}),
        ("pop context", {}),
        ("pop context", {}),
    ]


def test_order_change(widget):
    """The order of context objects can be changed."""
    # Initially nothing on the context.
    assert len(widget.context) == 0

    # Set up an inner context that has contained operations, including a subcontext
    widget.context.line_to(0, 0)
    with widget.Context() as context:
        context.line_to(10, 20)
        second = context.line_to(20, 30)
        with context.Fill() as fill:
            fill.line_to(25, 25)
        context.line_to(30, 40)
        context.line_to(40, 50)
    widget.context.line_to(99, 99)

    # Counts are as expected
    assert len(widget.context) == 3
    assert len(context) == 5
    assert len(fill) == 1

    # Initial draw instructions are as expected
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("line to", {"x": 10, "y": 20}),
        ("line to", {"x": 20, "y": 30}),
        # Begin fill
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 25,
                "y": 25,
                "fill_color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("fill", {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("pop context", {}),
        # End fill
        ("line to", {"x": 30, "y": 40}),
        ("line to", {"x": 40, "y": 50}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]

    # Remove the second draw instruction
    context.remove(second)

    # Drawing objects are as expected
    assert len(widget.context) == 3
    for i, cls in enumerate([LineTo, Context, LineTo]):
        assert isinstance(widget.context[i], cls)
    with pytest.raises(IndexError):
        widget.context[3]

    assert len(context) == 4
    assert len(fill) == 1

    # Draw instructions no longer have the second
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("line to", {"x": 10, "y": 20}),
        # Begin fill
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 25,
                "y": 25,
                "fill_color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("fill", {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("pop context", {}),
        # End fill
        ("line to", {"x": 30, "y": 40}),
        ("line to", {"x": 40, "y": 50}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]

    # Insert the second draw instruction at index 3
    context.insert(3, second)

    # Counts are as expected
    assert len(widget.context) == 3
    assert len(context) == 5
    assert len(fill) == 1

    # Draw instructions show the new position
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("line to", {"x": 10, "y": 20}),
        # Begin fill
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 25,
                "y": 25,
                "fill_color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("fill", {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("pop context", {}),
        # End fill
        ("line to", {"x": 30, "y": 40}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 40, "y": 50}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]

    # Remove the fill context
    context.remove(fill)

    # Counts are as expected
    assert len(widget.context) == 3
    assert len(context) == 4
    assert len(fill) == 1

    # Draw instructions show the new position
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("line to", {"x": 10, "y": 20}),
        ("line to", {"x": 30, "y": 40}),
        ("line to", {"x": 20, "y": 30}),
        ("line to", {"x": 40, "y": 50}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]

    # Insert the fill context at a negative index
    context.insert(-1, fill)
    # Draw instructions show the new position
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("line to", {"x": 10, "y": 20}),
        ("line to", {"x": 30, "y": 40}),
        ("line to", {"x": 20, "y": 30}),
        # Begin fill
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 25,
                "y": 25,
                "fill_color": rgb(0, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("fill", {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("pop context", {}),
        # End fill
        ("line to", {"x": 40, "y": 50}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]

    # Clear the context
    context.clear()

    # Counts are as expected
    assert len(widget.context) == 3
    assert len(context) == 0

    # No draw instructions other than the outer context.
    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        ("push context", {}),
        ("pop context", {}),
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]


def test_stacked_kwargs(widget):
    """If contexts are stacked, kwargs for sub operations don't leak."""
    widget.context.line_to(0, 0)
    with widget.Fill(color=rgb(255, 0, 0)) as fill1:
        fill1.line_to(10, 20)
        with fill1.Stroke(color=rgb(0, 255, 0)) as stroke1:
            stroke1.line_to(20, 30)
            with stroke1.Fill(color=rgb(0, 0, 255)) as fill2:
                fill2.line_to(100, 200)
                with fill2.Stroke(color=rgb(255, 255, 0)) as stroke2:
                    stroke2.line_to(200, 300)
                fill2.line_to(300, 400)
            stroke1.line_to(70, 80)
        fill1.line_to(80, 90)
    widget.context.line_to(99, 99)

    assert widget._impl.draw_instructions == [
        ("push context", {}),
        ("line to", {"x": 0, "y": 0}),
        # begin fill1
        ("push context", {}),
        ("begin path", {}),
        (
            "line to",
            {
                "x": 10,
                "y": 20,
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        # begin stroke 1
        ("push context", {"fill_color": rgb(255, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("begin path", {"fill_color": rgb(255, 0, 0), "fill_rule": FillRule.NONZERO}),
        (
            "line to",
            {
                "x": 20,
                "y": 30,
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # begin fill 2
        (
            "push context",
            {
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "begin path",
            {
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "line to",
            {
                "x": 100,
                "y": 200,
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # stroke 2
        (
            "push context",
            {
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "begin path",
            {
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "line to",
            {
                "x": 200,
                "y": 300,
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(255, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "stroke",
            {
                "color": rgb(255, 255, 0),
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "pop context",
            {
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # # end stroke 2
        (
            "line to",
            {
                "x": 300,
                "y": 400,
                "fill_color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "fill",
            {
                "color": rgb(0, 0, 255),
                "fill_rule": FillRule.NONZERO,
                "fill_color": rgb(255, 0, 0),
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "pop context",
            {
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        # end fill 2
        (
            "line to",
            {
                "x": 70,
                "y": 80,
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
                "stroke_color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
            },
        ),
        (
            "stroke",
            {
                "color": rgb(0, 255, 0),
                "line_width": 2.0,
                "line_dash": None,
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("pop context", {"fill_color": rgb(255, 0, 0), "fill_rule": FillRule.NONZERO}),
        # end stroke 1
        (
            "line to",
            {
                "x": 80,
                "y": 90,
                "fill_color": rgb(255, 0, 0),
                "fill_rule": FillRule.NONZERO,
            },
        ),
        ("fill", {"color": rgb(255, 0, 0), "fill_rule": FillRule.NONZERO}),
        ("pop context", {}),
        # end fill 1
        ("line to", {"x": 99, "y": 99}),
        ("pop context", {}),
    ]


def test_deprecated_args(widget):
    """Deprecated arguments to canvas functions raise warnings."""

    # fill() raises a warning about preserve being deprecated, then raises an error when
    # it's used as a context manager.
    with pytest.raises(
        RuntimeError,
        match=r"Context\.fill\(\) has been renamed Context\.Fill\(\)\.",
    ):
        with pytest.warns(
            DeprecationWarning,
            match=r"The `preserve` argument on fill\(\) has been deprecated.",
        ):
            with widget.context.fill(
                "rebeccapurple", FillRule.EVENODD, preserve=False
            ) as fill:
                fill.line_to(20, 30)
                fill.line_to(30, 20)



import pytest

import toga

from .test_command import assert_order


def test_create():
    """A group can be created with defaults."""
    grp = toga.Group("Group name")
    assert grp.text == "Group name"
    assert grp.order == 0

    assert repr(grp) == "<Group text='Group name' order=0>"


def test_create_with_params():
    """A fully specified group can be created."""
    parent = toga.Group("Parent name")
    grp = toga.Group("Group name", order=2, section=3, parent=parent)

    assert grp.text == "Group name"
    assert grp.order == 2
    assert grp.section == 3
    assert grp.parent == parent

    assert (
        repr(grp)
        == "<Group text='Group name' order=2 parent=<Group text='Parent name' order=0> section=3>"
    )


def test_create_section_without_parent():
    """A group cannot be created with a section but no parent."""
    with pytest.raises(
        ValueError,
        match=r"Section cannot be set without parent group",
    ):
        toga.Group("Group name", order=2, section=3)


def test_hashable():
    """Groups are hashable."""
    grp1 = toga.Group("text 1")
    grp2 = toga.Group("text 2")

    # The hash is based on the full path, not just the text.
    # This allows texts to be non-unique, as long as they're in
    # different groups
    grp1_child = toga.Group("text", parent=grp1)
    grp2_child = toga.Group("text", parent=grp2)

    # Insert the groups as keys in a dict. This is
    # only possible if Group is hashable.
    groups = {
        grp1: "First",
        grp2: "Second",
        grp1_child: "Child of 1",
        grp2_child: "Child of 2",
    }

    assert groups[grp1] == "First"
    assert groups[grp2] == "Second"
    assert groups[grp1_child] == "Child of 1"
    assert groups[grp2_child] == "Child of 2"


def test_group_eq():
    """Groups can be compared for equality."""
    group_a = toga.Group("A")
    group_b = toga.Group("B")
    group_a1 = toga.Group("A", order=1)
    # Assign None to variable to trick flake8 into not giving an E711
    other = None

    # Same instance is equal
    assert group_a == group_a
    assert group_a1 == group_a1

    # Same values are equal
    assert group_a == toga.Group("A")
    assert group_a1 == toga.Group("A", order=1)

    # Different values are not equal
    assert group_a != group_b
    assert group_a != other

    # Partially same values are not equal
    assert group_a1 != group_a
    assert group_a1 != toga.Group("B", order=1)
    assert group_a1 != toga.Group("A", order=2)


def test_parent_creation():
    """Parents can be assigned at creation."""
    group_a = toga.Group("A")
    group_b = toga.Group("B", parent=group_a)
    group_c = toga.Group("C", parent=group_b)

    # None checks
    assert not group_a.is_parent_of(None)
    assert not group_a.is_child_of(None)

    # Parent relationships
    assert group_a.is_parent_of(group_b)
    assert group_b.is_parent_of(group_c)
    assert group_a.is_parent_of(group_c)  # grandparent

    # Child relationships
    assert group_b.is_child_of(group_a)
    assert group_c.is_child_of(group_b)
    assert group_c.is_child_of(group_a)  # grandchild

    # Reverse direction relationships aren't true
    assert not group_a.is_child_of(group_b)
    assert not group_b.is_child_of(group_c)
    assert not group_a.is_child_of(group_c)

    assert not group_b.is_parent_of(group_a)
    assert not group_c.is_parent_of(group_b)
    assert not group_c.is_parent_of(group_a)

    assert group_a.parent is None
    assert group_a.root == group_a

    assert group_b.parent == group_a
    assert group_b.root == group_a

    assert group_c.parent == group_b
    assert group_c.root == group_a


def test_parent_assignment():
    """Parents can be assigned at runtime."""
    # Eventually, we'll end up with A->B->C, D.
    group_a = toga.Group("A")
    group_b = toga.Group("B")
    group_c = toga.Group("C")
    group_d = toga.Group("D")

    assert not group_a.is_parent_of(group_b)
    assert not group_b.is_parent_of(group_c)
    assert not group_a.is_parent_of(group_c)
    assert not group_b.is_parent_of(group_d)
    assert not group_a.is_parent_of(group_d)

    assert not group_b.is_child_of(group_a)
    assert not group_c.is_child_of(group_b)
    assert not group_c.is_child_of(group_a)
    assert not group_d.is_child_of(group_b)
    assert not group_d.is_child_of(group_a)

    assert not group_a.is_child_of(group_b)
    assert not group_b.is_child_of(group_c)
    assert not group_a.is_child_of(group_c)
    assert not group_b.is_child_of(group_d)
    assert not group_a.is_child_of(group_d)

    assert not group_b.is_parent_of(group_a)
    assert not group_c.is_parent_of(group_b)
    assert not group_c.is_parent_of(group_a)
    assert not group_d.is_parent_of(group_b)
    assert not group_d.is_parent_of(group_a)

    assert group_a.parent is None
    assert group_a.root == group_a

    assert group_b.parent is None
    assert group_b.root == group_b

    assert group_c.parent is None
    assert group_c.root == group_c

    # Assign parents.
    # C is assigned to B *before* B is assigned to A.
    # D is assigned to B *after* B is assigned to A.
    # This ensures that root isn't preserved in an intermediate state
    group_c.parent = group_b
    group_b.parent = group_a
    group_d.parent = group_b

    assert group_a.is_parent_of(group_b)
    assert group_b.is_parent_of(group_c)
    assert group_a.is_parent_of(group_c)  # grandparent
    assert group_b.is_parent_of(group_d)
    assert group_a.is_parent_of(group_d)  # grandparent

    assert group_b.is_child_of(group_a)
    assert group_c.is_child_of(group_b)
    assert group_c.is_child_of(group_a)  # grandchild
    assert group_d.is_child_of(group_b)
    assert group_d.is_child_of(group_a)  # grandchild

    # Reverse direction relationships aren't true
    assert not group_a.is_child_of(group_b)
    assert not group_b.is_child_of(group_c)
    assert not group_a.is_child_of(group_c)
    assert not group_b.is_child_of(group_d)
    assert not group_a.is_child_of(group_d)

    assert not group_b.is_parent_of(group_a)
    assert not group_c.is_parent_of(group_b)
    assert not group_c.is_parent_of(group_a)
    assert not group_d.is_parent_of(group_b)
    assert not group_d.is_parent_of(group_a)

    assert group_a.parent is None
    assert group_a.root == group_a

    assert group_b.parent == group_a
    assert group_b.root == group_a

    assert group_c.parent == group_b
    assert group_c.root == group_a

    assert group_d.parent == group_b
    assert group_d.root == group_a


def test_parent_loops():
    """Parent loops are prevented can be assigned at runtime."""
    group_a = toga.Group("A")
    group_b = toga.Group("B", parent=group_a)
    group_c = toga.Group("C", parent=group_b)

    #
    with pytest.raises(
        ValueError,
        match=r"A group cannot be it's own parent",
    ):
        group_a.parent = group_a

    with pytest.raises(
        ValueError,
        match=r"Cannot set parent; 'A' is an ancestor of 'B'.",
    ):
        group_a.parent = group_b

    with pytest.raises(
        ValueError,
        match=r"Cannot set parent; 'A' is an ancestor of 'C'.",
    ):
        group_a.parent = group_c


def test_order_by_text():
    """Groups are ordered by text if order and section are equivalent."""
    assert_order(toga.Group("A"), toga.Group("B"))


def test_order_by_number():
    """Groups are ordered by number."""
    assert_order(toga.Group("B", order=1), toga.Group("A", order=2))


def test_order_by_groups(parent_group_1, parent_group_2):
    """Groups are ordered by parent, then section, then order."""
    assert_order(
        parent_group_1,
        toga.Group("C", parent=parent_group_1),
        toga.Group("D", parent=parent_group_1),
        toga.Group("AA3", parent=parent_group_1, section=2),
        toga.Group("AA2", parent=parent_group_1, section=3, order=1),
        toga.Group("AA1", parent=parent_group_1, section=3, order=2),
        parent_group_2,
        toga.Group("B", parent=parent_group_2),
    )


# def test_cannot_set_self_as_parent(self):
#     group = toga.Group("P")
#     with self.assertRaises(ValueError):
#         group.parent = group
#     self.assert_parent_and_root(group, None, group)


# def test_cannot_set_child_to_be_a_parent_of_its_grandparent(self):
#     grandparent = toga.Group("G")
#     parent = toga.Group("P", parent=grandparent)
#     child = toga.Group("C", parent=parent)
#     with self.assertRaises(ValueError):
#         grandparent.parent = child
#     self.assert_parent_and_root(grandparent, None, grandparent)
#     self.assert_parent_and_root(parent, grandparent, grandparent)
#     self.assert_parent_and_root(child, parent, grandparent)


# def assert_parent_and_root(self, group, parent, root):
#     self.assertEqual(group.parent, parent)
#     self.assertEqual(group.root, root)



import sys

import pytest

import toga
from toga_dummy.utils import EventLog


@pytest.fixture(autouse=True)
def reset_event_log():
    EventLog.reset()


@pytest.fixture(autouse=True)
def clear_sys_modules(monkeypatch):
    try:
        # App startup is influenced by things like the state of sys.modules, and the
        # presence of __main__ in particular. Pytest doesn't need __main__ to work;
        # so if it exists, delete it for the purposes of each test.
        monkeypatch.delitem(sys.modules, "__main__")
    except KeyError:
        pass


class TestApp(toga.App):
    pass


@pytest.fixture
def app(event_loop):
    # The app icon is cached; purge the app icon cache if it exists
    try:
        del toga.Icon.__APP_ICON
    except AttributeError:
        pass

    return TestApp(formal_name="Test App", app_id="org.beeware.toga.test-app")



from datetime import date, datetime, time
from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import assert_action_performed


@pytest.fixture
def on_change_handler():
    return Mock()


@pytest.fixture
def widget(on_change_handler):
    return toga.DateInput(on_change=on_change_handler)


@pytest.mark.freeze_time("2023-05-25")
def test_widget_created():
    """A DateInput can be created."""
    widget = toga.DateInput()

    # Round trip the impl/interface
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create DateInput")

    assert widget.value == date(2023, 5, 25)
    assert widget.min == date(1800, 1, 1)
    assert widget.max == date(8999, 12, 31)
    assert widget.on_change._raw is None


def test_widget_created_with_values(on_change_handler):
    """A DateInput can be created with initial values."""
    # Round trip the impl/interface
    widget = toga.DateInput(
        value=date(2015, 6, 15),
        min=date(2013, 5, 14),
        max=date(2017, 7, 16),
        on_change=on_change_handler,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create DateInput")

    assert widget.value == date(2015, 6, 15)
    assert widget.min == date(2013, 5, 14)
    assert widget.max == date(2017, 7, 16)
    assert widget.on_change._raw == on_change_handler

    # The change handler isn't invoked at construction.
    on_change_handler.assert_not_called()


@pytest.mark.freeze_time("2023-05-25")
@pytest.mark.parametrize(
    "value, expected",
    [
        (date(2023, 1, 11), date(2023, 1, 11)),
        (datetime(2023, 2, 11, 10, 42, 37), date(2023, 2, 11)),
        ("2023-03-11", date(2023, 3, 11)),
        (None, date(2023, 5, 25)),
    ],
)
def test_value(widget, value, expected, on_change_handler):
    """The value of the datepicker can be set."""
    widget.value = value

    assert widget.value == expected

    on_change_handler.assert_called_once_with(widget)


INVALID_VALUES = [
    (123, TypeError, "Not a valid date value"),
    (object(), TypeError, "Not a valid date value"),
    (time(10, 42, 37), TypeError, "Not a valid date value"),
    ("not a date", ValueError, "Invalid isoformat string: 'not a date'"),
]


@pytest.mark.parametrize("value, exc, message", INVALID_VALUES)
def test_invalid_value(widget, value, exc, message):
    """Invalid date values raise an exception."""
    with pytest.raises(exc, match=message):
        widget.value = value


@pytest.mark.parametrize(
    "value, clipped",
    [
        (date(2005, 6, 12), date(2010, 1, 1)),
        (date(2009, 12, 31), date(2010, 1, 1)),
        (date(2010, 1, 1), date(2010, 1, 1)),
        (date(2015, 6, 12), date(2015, 6, 12)),
        (date(2020, 1, 1), date(2020, 1, 1)),
        (date(2020, 1, 2), date(2020, 1, 1)),
        (date(2023, 6, 12), date(2020, 1, 1)),
        # Unlike `min` and `max`, `value` accepts and clips dates outside of the
        # supported range.
        (date(1700, 1, 1), date(2010, 1, 1)),
        (date(9999, 12, 31), date(2020, 1, 1)),
    ],
)
def test_value_clipping(widget, value, clipped, on_change_handler):
    """It the value is inconsistent with min/max, it is clipped."""
    # Set min/max dates, and clear the on_change mock
    widget.min = date(2010, 1, 1)
    widget.max = date(2020, 1, 1)
    on_change_handler.reset_mock()

    # Set the new value
    widget.value = value

    # Value has been clipped
    assert widget.value == clipped

    # on_change handler called once.
    on_change_handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, date(1800, 1, 1)),
        (date(2023, 1, 11), date(2023, 1, 11)),
        (datetime(2023, 2, 11, 10, 42, 37), date(2023, 2, 11)),
        ("2023-03-11", date(2023, 3, 11)),
    ],
)
def test_min(widget, value, expected):
    """The min of the datepicker can be set."""
    widget.min = value

    assert widget.min == expected


INVALID_LIMITS = INVALID_VALUES + [
    (date(1799, 12, 31), ValueError, "The lowest supported date is 1800-01-01"),
    (date(9000, 1, 1), ValueError, "The highest supported date is 8999-12-31"),
]


@pytest.mark.parametrize("value, exc, message", INVALID_LIMITS)
def test_invalid_min(widget, value, exc, message):
    """Invalid min values raise an exception."""
    widget.max = date(2025, 6, 12)

    with pytest.raises(exc, match=message):
        widget.min = value


@pytest.mark.parametrize(
    "min, clip_value, clip_max",
    [
        (date(2005, 6, 1), False, False),
        (date(2005, 6, 25), False, False),
        (date(2005, 6, 26), True, False),
        (date(2005, 7, 4), True, False),
        (date(2005, 12, 31), True, False),
        (date(2006, 1, 1), True, True),
        (date(2006, 7, 4), True, True),
    ],
)
def test_min_clip(widget, on_change_handler, min, clip_value, clip_max):
    """If the current value or max is before a new min date, it is clipped."""
    widget.value = date(2005, 6, 25)
    widget.max = date(2005, 12, 31)
    on_change_handler.reset_mock()

    widget.min = min
    assert widget.min == min

    if clip_value:
        assert widget.value == min
        on_change_handler.assert_called_once_with(widget)
    else:
        assert widget.value == date(2005, 6, 25)
        on_change_handler.assert_not_called()

    if clip_max:
        assert widget.max == min
    else:
        assert widget.max == date(2005, 12, 31)


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, date(8999, 12, 31)),
        (date(2023, 1, 11), date(2023, 1, 11)),
        (datetime(2023, 2, 11, 10, 42, 37), date(2023, 2, 11)),
        ("2023-03-11", date(2023, 3, 11)),
    ],
)
def test_max(widget, value, expected):
    """The max of the datepicker can be set."""
    widget.max = value

    assert widget.max == expected


@pytest.mark.parametrize("value, exc, message", INVALID_LIMITS)
def test_invalid_max(widget, value, exc, message):
    """Invalid max values raise an exception."""
    widget.min = date(2015, 6, 12)

    with pytest.raises(exc, match=message):
        widget.max = value


@pytest.mark.parametrize(
    "max, clip_value, clip_min",
    [
        (date(2005, 6, 1), True, True),
        (date(2005, 6, 24), True, True),
        (date(2005, 6, 25), True, False),
        (date(2005, 7, 4), True, False),
        (date(2005, 12, 30), True, False),
        (date(2005, 12, 31), False, False),
        (date(2006, 7, 4), False, False),
    ],
)
def test_max_clip(widget, on_change_handler, max, clip_value, clip_min):
    """If the current value or min is after a new max date, it is clipped."""
    widget.min = date(2005, 6, 25)
    widget.value = date(2005, 12, 31)
    on_change_handler.reset_mock()

    widget.max = max
    assert widget.max == max

    if clip_value:
        assert widget.value == max
        on_change_handler.assert_called_once_with(widget)
    else:
        assert widget.value == date(2005, 12, 31)
        on_change_handler.assert_not_called()

    if clip_min:
        assert widget.min == max
    else:
        assert widget.min == date(2005, 6, 25)


def test_deprecated_names():
    MIN = date(2012, 8, 3)
    MAX = date(2016, 11, 15)

    with pytest.warns(
        DeprecationWarning, match="DatePicker has been renamed DateInput"
    ), pytest.warns(
        DeprecationWarning, match="DatePicker.min_date has been renamed DateInput.min"
    ), pytest.warns(
        DeprecationWarning, match="DatePicker.max_date has been renamed DateInput.max"
    ):
        widget = toga.DatePicker(min_date=MIN, max_date=MAX)

    assert widget.min == MIN
    assert widget.max == MAX
    widget.min = widget.max = None

    with pytest.warns(
        DeprecationWarning, match="DatePicker.min_date has been renamed DateInput.min"
    ):
        widget.min_date = MIN

    with pytest.warns(
        DeprecationWarning, match="DatePicker.min_date has been renamed DateInput.min"
    ):
        assert widget.min_date == MIN
    assert widget.min == MIN

    with pytest.warns(
        DeprecationWarning, match="DatePicker.max_date has been renamed DateInput.max"
    ):
        widget.max_date = MAX

    with pytest.warns(
        DeprecationWarning, match="DatePicker.max_date has been renamed DateInput.max"
    ):
        assert widget.max_date == MAX
    assert widget.max == MAX

    with pytest.warns(
        DeprecationWarning, match="DatePicker has been renamed DateInput"
    ):
        widget = toga.DatePicker()
    assert widget.min == date(1800, 1, 1)
    assert widget.max == date(8999, 12, 31)



from __future__ import annotations

import datetime
import warnings
from typing import Any, Protocol

import toga
from toga.handlers import wrapped_handler

from .base import StyleT, Widget


class OnChangeHandler(Protocol):
    def __call__(self, widget: TimeInput, /, **kwargs: Any) -> object:
        """A handler to invoke when the time input is changed.

        :param widget: The TimeInput that was changed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class TimeInput(Widget):
    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        value: datetime.time | None = None,
        min: datetime.time | None = None,
        max: datetime.time | None = None,
        on_change: toga.widgets.timeinput.OnChangeHandler | None = None,
    ):
        """Create a new TimeInput widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param value: The initial time to display. If not specified, the current time
            will be used.
        :param min: The earliest time (inclusive) that can be selected.
        :param max: The latest time (inclusive) that can be selected.
        :param on_change: A handler that will be invoked when the value changes.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a TimeInput
        self._impl = self.factory.TimeInput(interface=self)

        self.on_change = None
        self.min = min
        self.max = max

        self.value = value
        self.on_change = on_change

    @property
    def value(self) -> datetime.time:
        """The currently selected time. A value of ``None`` will be converted into the
        current time.

        If this property is set to a value outside of the min/max range, it will be
        clipped.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value: object) -> None:
        value = self._convert_time(value)

        if value < self.min:
            value = self.min
        elif value > self.max:
            value = self.max

        self._impl.set_value(value)

    def _convert_time(self, value: object) -> datetime.time:
        if value is None:
            value = datetime.datetime.now().time()
        elif isinstance(value, datetime.datetime):
            value = value.time()
        elif isinstance(value, datetime.time):
            pass
        elif isinstance(value, str):
            value = datetime.time.fromisoformat(value)
        else:
            raise TypeError("Not a valid time value")

        return value.replace(microsecond=0)

    @property
    def min(self) -> datetime.time:
        """The minimum allowable time (inclusive). A value of ``None`` will be converted
        into 00:00:00.

        When setting this property, the current :attr:`value` and :attr:`max` will be
        clipped against the new minimum value.
        """
        return self._impl.get_min_time()

    @min.setter
    def min(self, value: object) -> None:
        if value is None:
            min = datetime.time(0, 0, 0)
        else:
            min = self._convert_time(value)

        if self.max < min:
            self._impl.set_max_time(min)
        self._impl.set_min_time(min)
        if self.value < min:
            self.value = min

    @property
    def max(self) -> datetime.time:
        """The maximum allowable time (inclusive). A value of ``None`` will be converted
        into 23:59:59.

        When setting this property, the current :attr:`value` and :attr:`min` will be
        clipped against the new maximum value.
        """
        return self._impl.get_max_time()

    @max.setter
    def max(self, value: object) -> None:
        if value is None:
            max = datetime.time(23, 59, 59)
        else:
            max = self._convert_time(value)

        if self.min > max:
            self._impl.set_min_time(max)
        self._impl.set_max_time(max)
        if self.value > max:
            self.value = max

    @property
    def on_change(self) -> OnChangeHandler:
        """The handler to invoke when the time value changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler: toga.widgets.timeinput.OnChangeHandler) -> None:
        self._on_change = wrapped_handler(self, handler)


# 2023-05: Backwards compatibility
class TimePicker(TimeInput):
    def __init__(self, *args: Any, **kwargs: Any):
        warnings.warn("TimePicker has been renamed TimeInput", DeprecationWarning)

        for old_name, new_name in [
            ("min_time", "min"),
            ("max_time", "max"),
        ]:
            try:
                value = kwargs.pop(old_name)
                warnings.warn(
                    f"TimePicker.{old_name} has been renamed TimeInput.{new_name}",
                    DeprecationWarning,
                )
            except KeyError:
                pass
            else:
                kwargs[new_name] = value

        super().__init__(*args, **kwargs)

    @property
    def min_time(self) -> datetime.time:
        warnings.warn(
            "TimePicker.min_time has been renamed TimeInput.min", DeprecationWarning
        )
        return self.min

    @min_time.setter
    def min_time(self, value: object) -> None:
        warnings.warn(
            "TimePicker.min_time has been renamed TimeInput.min", DeprecationWarning
        )
        self.min = value

    @property
    def max_time(self) -> datetime.time:
        warnings.warn(
            "TimePicker.max_time has been renamed TimeInput.max", DeprecationWarning
        )
        return self.max

    @max_time.setter
    def max_time(self, value: object) -> None:
        warnings.warn(
            "TimePicker.max_time has been renamed TimeInput.max", DeprecationWarning
        )
        self.max = value



from unittest.mock import Mock

from toga.sources.value_source import ValueSource


def test_empty():
    """An empty ValueSource can be created."""
    source = ValueSource()

    assert source.accessor == "value"
    assert source.value is None
    assert str(source) == "None"


def test_value():
    """A ValueSource can be created with a value."""
    source = ValueSource(42)

    assert source.accessor == "value"
    assert source.value == 42
    assert str(source) == "42"


def test_value_with_accessor():
    """A ValueSource can be created with a custom accessor."""
    source = ValueSource(42, accessor="something")

    assert source.accessor == "something"
    assert source.something == 42
    assert str(source) == "42"


def test_listener():
    """A listener will be notified of ValueSource changes."""
    source = ValueSource(42)
    listener = Mock()

    source.add_listener(listener)

    source.value = 37

    listener.change.assert_called_once_with(item=37)

    # Reset the mock; clear the value
    listener.reset_mock()
    source.value = None

    listener.change.assert_called_once_with(item=None)

    # Reset the mock; set a *different* attribute on the source
    listener.reset_mock()
    source.something = 1234

    listener.change.assert_not_called()



import toga


class SubclassedApp(toga.App):
    pass


def main():
    app = SubclassedApp("Subclassed App", "org.testbed.subclassed-app")

    print(f"app.paths.app={app.paths.app.resolve()}")
    print(f"app.paths.config={app.paths.config.resolve()}")
    print(f"app.paths.data={app.paths.data.resolve()}")
    print(f"app.paths.cache={app.paths.cache.resolve()}")
    print(f"app.paths.logs={app.paths.logs.resolve()}")
    print(f"app.paths.toga={app.paths.toga.resolve()}")


if __name__ == "__main__":
    main()



from toga.keys import Key


def test_is_printable():
    """Key printability can be checked."""
    assert not Key.is_printable(Key.SHIFT)
    assert Key.is_printable(Key.LESS_THAN)
    assert Key.is_printable(Key.GREATER_THAN)
    assert Key.is_printable(Key.NUMPAD_0)


def test_modifiers():
    """Keys can be added with modifiers."""
    # Mod + Key
    assert Key.MOD_1 + Key.A == "<mod 1>a"

    # Multiple modifiers can be used
    assert Key.MOD_1 + Key.SHIFT + Key.A == "<mod 1><shift>a"

    # Bare characters can be used
    assert Key.MOD_1 + "a" == "<mod 1>a"



import asyncio
from unittest.mock import Mock

import pytest

from toga.handlers import AsyncResult, NativeHandler, wrapped_handler


class ExampleAsyncResult(AsyncResult):
    RESULT_TYPE = "Test"


def test_noop_handler():
    """None can be wrapped as a valid handler."""
    obj = Mock()

    wrapped = wrapped_handler(obj, None)

    assert wrapped._raw is None

    # This does nothing, but doesn't raise an error.
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)


def test_noop_handler_with_cleanup():
    """Cleanup is still performed when a no-op handler is used."""
    obj = Mock()
    cleanup = Mock()

    wrapped = wrapped_handler(obj, None, cleanup=cleanup)

    assert wrapped._raw is None

    # This does nothing, but doesn't raise an error.
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, None)


def test_noop_handler_with_cleanup_error(capsys):
    """If cleanup on a no-op handler raises an error, it is logged."""
    obj = Mock()
    cleanup = Mock(side_effect=Exception("Problem in cleanup"))

    wrapped = wrapped_handler(obj, None, cleanup=cleanup)

    assert wrapped._raw is None

    # This does nothing, but doesn't raise an error.
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, None)

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in handler cleanup: Problem in cleanup\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_function_handler():
    """A function can be used as a handler."""
    obj = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs

    wrapped = wrapped_handler(obj, handler)

    # Raw handler is the original function
    assert wrapped._raw == handler

    # Invoke wrapper
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }


def test_function_handler_error(capsys):
    """A function handler can raise an error."""
    obj = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        raise Exception("Problem in handler")

    wrapped = wrapped_handler(obj, handler)

    assert wrapped._raw == handler

    # Invoke handler. The exception is swallowed
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }

    # Evidence of the handler error is in the log.
    assert (
        "Error in handler: Problem in handler\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_function_handler_with_cleanup():
    """A function handler can have a cleanup method."""
    obj = Mock()
    cleanup = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original function
    assert wrapped._raw == handler

    # Invoke handler
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)


def test_function_handler_with_cleanup_error(capsys):
    """A function handler can have a cleanup method that raises an error."""
    obj = Mock()
    cleanup = Mock(side_effect=Exception("Problem in cleanup"))
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original function
    assert wrapped._raw == handler

    # Invoke handler. The exception in cleanup is swallowed
    wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in handler cleanup: Problem in cleanup\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_generator_handler(event_loop):
    """A generator can be used as a handler."""
    obj = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        yield 0.01  # A short sleep
        handler_call["slept"] = True
        yield  # A yield without a sleep
        handler_call["done"] = True

    wrapped = wrapped_handler(obj, handler)

    # Raw handler is the original generator
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "slept": True,
        "done": True,
    }


def test_generator_handler_error(event_loop, capsys):
    """A generator can raise an error."""
    obj = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        yield 0.01  # A short sleep
        raise Exception("Problem in handler")

    wrapped = wrapped_handler(obj, handler)

    # Raw handler is the original generator
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in long running handler: Problem in handler\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_generator_handler_with_cleanup(event_loop):
    """A generator can have cleanup."""
    obj = Mock()
    cleanup = Mock()
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        yield 0.01  # A short sleep
        handler_call["slept"] = True
        yield  # A yield without a sleep
        handler_call["done"] = True
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original generator
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "slept": True,
        "done": True,
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)


def test_generator_handler_with_cleanup_error(event_loop, capsys):
    """A generator can raise an error during cleanup."""
    obj = Mock()
    cleanup = Mock(side_effect=Exception("Problem in cleanup"))
    handler_call = {}

    def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        yield 0.01  # A short sleep
        handler_call["slept"] = True
        yield  # A yield without a sleep
        handler_call["done"] = True
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original generator
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "slept": True,
        "done": True,
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in long running handler cleanup: Problem in cleanup\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_coroutine_handler(event_loop):
    """A coroutine can be used as a handler."""
    obj = Mock()
    handler_call = {}

    async def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        await asyncio.sleep(0.01)  # A short sleep
        handler_call["done"] = True

    wrapped = wrapped_handler(obj, handler)

    # Raw handler is the original coroutine
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "done": True,
    }


def test_coroutine_handler_error(event_loop, capsys):
    """A coroutine can raise an error."""
    obj = Mock()
    handler_call = {}

    async def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        await asyncio.sleep(0.01)  # A short sleep
        raise Exception("Problem in handler")

    wrapped = wrapped_handler(obj, handler)

    # Raw handler is the original coroutine
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
    }

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in async handler: Problem in handler\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_coroutine_handler_with_cleanup(event_loop):
    """A coroutine can have cleanup."""
    obj = Mock()
    cleanup = Mock()
    handler_call = {}

    async def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        await asyncio.sleep(0.01)  # A short sleep
        handler_call["done"] = True
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original coroutine
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "done": True,
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)


def test_coroutine_handler_with_cleanup_error(event_loop, capsys):
    """A coroutine can raise an error during cleanup."""
    obj = Mock()
    cleanup = Mock(side_effect=Exception("Problem in cleanup"))
    handler_call = {}

    async def handler(*args, **kwargs):
        handler_call["args"] = args
        handler_call["kwargs"] = kwargs
        await asyncio.sleep(0.01)  # A short sleep
        handler_call["done"] = True
        return 42

    wrapped = wrapped_handler(obj, handler, cleanup=cleanup)

    # Raw handler is the original coroutine
    assert wrapped._raw == handler

    # Invoke wrapper inside an active run loop.
    async def waiter():
        wrapped("arg1", "arg2", kwarg1=3, kwarg2=4)
        count = 0
        while not handler_call.get("done", False) and count < 5:
            await asyncio.sleep(0.01)
            count += 1

    event_loop.run_until_complete(waiter())

    # Handler arguments are as expected.
    assert handler_call == {
        "args": (obj, "arg1", "arg2"),
        "kwargs": {"kwarg1": 3, "kwarg2": 4},
        "done": True,
    }

    # Cleanup method was invoked
    cleanup.assert_called_once_with(obj, 42)

    # Evidence of the handler cleanup error is in the log.
    assert (
        "Error in async handler cleanup: Problem in cleanup\nTraceback (most recent call last):\n"
        in capsys.readouterr().err
    )


def test_native_handler():
    """A native function can be used as a handler."""
    obj = Mock()
    native_method = Mock()

    handler = NativeHandler(native_method)

    wrapped = wrapped_handler(obj, handler)

    # Native method is returned as the handler.
    assert wrapped == native_method


def test_async_result_non_comparable(event_loop):
    """An async result can't be compared or evaluated."""
    result = ExampleAsyncResult(None)

    # repr for the result is useful
    assert repr(result) == "<Async Test result; future=<Future pending>>"

    # Result cannot be compared.

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result == 42

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result < 42

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result <= 42

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result > 42

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result >= 42

    with pytest.raises(
        RuntimeError,
        match=r"Can't check Test result directly; use await or an on_result handler",
    ):
        result != 42


def test_async_result(event_loop):
    """An async result can be set."""
    result = ExampleAsyncResult()

    result.set_result(42)

    # Evaluate the future
    async_answer = event_loop.run_until_complete(result.future)

    # The answer was returned, and passed to the callback
    assert async_answer == 42


def test_async_result_cancelled(event_loop):
    """An async result can be set even if the future is cancelled."""
    result = ExampleAsyncResult()

    # cancel the future.
    result.future.cancel()

    result.set_result(42)

    # Evaluate the future. This will raise an error
    with pytest.raises(asyncio.CancelledError):
        event_loop.run_until_complete(result.future)


def test_async_exception(event_loop):
    """An async result can raise an exception."""
    result = ExampleAsyncResult()

    result.set_exception(ValueError("Bad stuff"))

    # Evaluate the future; this will raise an exception
    with pytest.raises(ValueError, match=r"Bad stuff"):
        event_loop.run_until_complete(result.future)


def test_async_exception_cancelled(event_loop):
    """An async result can raise an exception even if the future is cancelled."""
    result = ExampleAsyncResult()

    # Cancel the future
    result.future.cancel()

    result.set_exception(ValueError("Bad stuff"))

    # Evaluate the future. This will raise an error
    with pytest.raises(asyncio.CancelledError):
        event_loop.run_until_complete(result.future)


######################################################################
# 2023-12: Backwards compatibility
######################################################################


def test_async_result_sync(event_loop):
    """The deprecated behavior of using a synchronous result handler is supported."""
    on_result = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        result = ExampleAsyncResult(on_result)

    result.set_result(42)

    # Evaluate the future
    async_answer = event_loop.run_until_complete(result.future)

    # The answer was returned, and passed to the callback
    assert async_answer == 42
    on_result.assert_called_once_with(42)


def test_async_result_cancelled_sync(event_loop):
    """A deprecated on_result handler won't be fired on a cancelled future."""
    on_result = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        result = ExampleAsyncResult(on_result)

    # cancel the future.
    result.future.cancel()

    result.set_result(42)

    # Evaluate the future. This will raise an error
    with pytest.raises(asyncio.CancelledError):
        event_loop.run_until_complete(result.future)

    # The callback wasn't called
    on_result.assert_not_called()


def test_async_exception_sync(event_loop):
    """A deprecated on_result handler can raise an exception."""
    on_result = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        result = ExampleAsyncResult(on_result)

    result.set_exception(ValueError("Bad stuff"))

    # Evaluate the future; this will raise an exception
    with pytest.raises(ValueError, match=r"Bad stuff"):
        event_loop.run_until_complete(result.future)

    # The answer was returned, and passed to the callback
    on_result.assert_called_once()
    assert on_result.call_args[0] == (None,)
    assert isinstance(on_result.call_args[1]["exception"], ValueError)


def test_async_exception_cancelled_sync(event_loop):
    """An async result can raise an exception even if the future is cancelled."""
    on_result = Mock()

    with pytest.warns(
        DeprecationWarning,
        match=r"Synchronous `on_result` handlers have been deprecated;",
    ):
        result = ExampleAsyncResult(on_result)

    # Cancel the future
    result.future.cancel()

    result.set_exception(ValueError("Bad stuff"))

    # Evaluate the future. This will raise an error
    with pytest.raises(asyncio.CancelledError):
        event_loop.run_until_complete(result.future)

    # The callback wasn't called
    on_result.assert_not_called()



from __future__ import annotations

import asyncio
from typing import Any, Protocol

from toga.handlers import AsyncResult, OnResultT, wrapped_handler

from .base import StyleT, Widget


class JavaScriptResult(AsyncResult):
    RESULT_TYPE = "JavaScript"


class OnWebViewLoadHandler(Protocol):
    def __call__(self, widget: WebView, /, **kwargs: Any) -> object:
        """A handler to invoke when the WebView is loaded.

        :param widget: The WebView that was loaded.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class WebView(Widget):
    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        url: str | None = None,
        user_agent: str | None = None,
        on_webview_load: OnWebViewLoadHandler | None = None,
    ):
        """Create a new WebView widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param url: The full URL to load in the WebView. If not provided,
            an empty page will be displayed.
        :param user_agent: The user agent to use for web requests. If not
            provided, the default user agent for the platform will be used.
        :param on_webview_load: A handler that will be invoked when the web view
            finishes loading.
        """
        super().__init__(id=id, style=style)

        self._impl = self.factory.WebView(interface=self)
        self.user_agent = user_agent

        # Set the load handler before loading the first URL.
        self.on_webview_load = on_webview_load
        self.url = url

    def _set_url(self, url: str | None, future: asyncio.Future | None) -> None:
        # Utility method for validating and setting the URL with a future.
        if (url is not None) and not url.startswith(("https://", "http://")):
            raise ValueError("WebView can only display http:// and https:// URLs")

        self._impl.set_url(url, future=future)

    @property
    def url(self) -> str | None:
        """The current URL, or ``None`` if no URL is currently displayed.

        After setting this property, it is not guaranteed that reading the property will
        immediately return the new value. To be notified once the URL has finished
        loading, use :any:`load_url` or :any:`on_webview_load`.
        """
        return self._impl.get_url()

    @url.setter
    def url(self, value: str | None) -> None:
        self._set_url(value, future=None)

    async def load_url(self, url: str) -> asyncio.Future:
        """Load a URL, and wait until the next :any:`on_webview_load` event.

        **Note:** On Android, this method will return immediately.

        :param url: The URL to load.
        """
        loop = asyncio.get_event_loop()
        loaded_future = loop.create_future()
        self._set_url(url, future=loaded_future)
        return await loaded_future

    @property
    def on_webview_load(self) -> OnWebViewLoadHandler:
        """The handler to invoke when the web view finishes loading.

        Rendering web content is a complex, multi-threaded process. Although a page
        may have completed *loading*, there's no guarantee that the page has been
        fully *rendered*, or that the widget representation has been fully updated.
        The number of load events generated by a URL transition or content change can
        be unpredictable. An ``on_webview_load`` event should be interpreted as an
        indication that some change has occurred, not that a *specific* change has
        occurred, or that a specific change has been fully propagated into the
        rendered content.

        **Note:** This is not currently supported on Android.
        """
        return self._on_webview_load

    @on_webview_load.setter
    def on_webview_load(self, handler: OnWebViewLoadHandler) -> None:
        if handler and not getattr(self._impl, "SUPPORTS_ON_WEBVIEW_LOAD", True):
            self.factory.not_implemented("WebView.on_webview_load")

        self._on_webview_load = wrapped_handler(self, handler)

    @property
    def user_agent(self) -> str:
        """The user agent to use for web requests.

        **Note:** On Windows, this property will return an empty string until the widget
        has finished initializing.
        """
        return self._impl.get_user_agent()

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self._impl.set_user_agent(value)

    def set_content(self, root_url: str, content: str) -> None:
        """Set the HTML content of the WebView.

        **Note:** On Android and Windows, the ``root_url`` argument is ignored. Calling
        this method will set the ``url`` property to ``None``.

        :param root_url: A URL which will be returned by the ``url`` property,
            and used to resolve any relative URLs in the content.
        :param content: The HTML content for the WebView
        """
        self._impl.set_content(root_url, content)

    def evaluate_javascript(
        self,
        javascript: str,
        on_result: OnResultT | None = None,
    ) -> JavaScriptResult:
        """Evaluate a JavaScript expression.

        **This is an asynchronous method**. There is no guarantee that the JavaScript
        has finished evaluating when this method returns. The object returned by this
        method can be awaited to obtain the value of the expression.

        **Note:** On Android and Windows, *no exception handling is performed*. If a
        JavaScript error occurs, a return value of None will be reported, but no
        exception will be provided.

        :param javascript: The JavaScript expression to evaluate.
        :param on_result: **DEPRECATED** ``await`` the return value of this method.
        """
        return self._impl.evaluate_javascript(javascript, on_result=on_result)



import asyncio
import importlib.metadata
import sys
import webbrowser
from pathlib import Path
from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)

EXPLICIT_FULL_APP_KWARGS = dict(
    formal_name="Explicit App",
    app_id="org.beeware.explicit-app",
    app_name="override-app",
)
EXPLICIT_MIN_APP_KWARGS = dict(
    formal_name="Explicit App",
    app_id="org.beeware.explicit-app",
)
APP_METADATA = {
    "Formal-Name": "Test App",
    "App-ID": "org.beeware.test-app",
    "Name": "test-app",
}


@pytest.mark.parametrize(
    (
        "kwargs, metadata, main_module, expected_formal_name, expected_app_id, "
        "expected_app_name"
    ),
    [
        ###########################################################################
        # Invoking as python my_app.py, or as an interactive prompt
        # This causes a main package of None
        ###########################################################################
        # Explicit app properties, no metadata
        (
            EXPLICIT_FULL_APP_KWARGS,
            None,
            Mock(__package__=None),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        # Explicit app properties, but implied distribution name from app_id, no
        # metadata
        (
            EXPLICIT_MIN_APP_KWARGS,
            None,
            Mock(__package__=None),
            "Explicit App",
            "org.beeware.explicit-app",
            "explicit-app",
        ),
        # No app properties, with metadata
        (
            dict(),
            APP_METADATA,
            Mock(__package__=None),
            "Test App",
            "org.beeware.test-app",
            "toga",
        ),
        # Explicit app properties, with metadata. Explicit values take precedence.
        (
            EXPLICIT_FULL_APP_KWARGS,
            APP_METADATA,
            Mock(__package__=None),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        ###########################################################################
        # Invoking as python -m my_app, where code is in my_app.py
        # This causes a main module of ""
        ###########################################################################
        # Explicit app properties, no metadata
        (
            EXPLICIT_FULL_APP_KWARGS,
            None,
            Mock(__package__=""),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        # Explicit app properties, but implied distribution name from app_id, no metadata
        (
            EXPLICIT_MIN_APP_KWARGS,
            None,
            Mock(__package__=""),
            "Explicit App",
            "org.beeware.explicit-app",
            "explicit-app",
        ),
        # No app properties, with metadata
        (
            dict(),
            APP_METADATA,
            Mock(__package__=""),
            "Test App",
            "org.beeware.test-app",
            "toga",
        ),
        # Explicit app properties, with metadata. Explicit values take precedence.
        (
            EXPLICIT_FULL_APP_KWARGS,
            APP_METADATA,
            Mock(__package__=""),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        ###########################################################################
        # Invoking as python -m my_app, where my_app is a folder with a __main__
        # This causes a main module of "my_app"
        ###########################################################################
        # Explicit app properties, no metadata
        (
            EXPLICIT_FULL_APP_KWARGS,
            None,
            Mock(__package__="my_app"),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        # Explicit app properties, but implied distribution name from __package__, no
        # metadata
        (
            EXPLICIT_MIN_APP_KWARGS,
            None,
            Mock(__package__="my_app"),
            "Explicit App",
            "org.beeware.explicit-app",
            "my_app",
        ),
        # No app properties, with metadata
        (
            dict(),
            APP_METADATA,
            Mock(__package__="my_app"),
            "Test App",
            "org.beeware.test-app",
            "my_app",
        ),
        # Explicit app properties, with metadata. Explicit values take precedence.
        (
            EXPLICIT_FULL_APP_KWARGS,
            APP_METADATA,
            Mock(__package__="my_app"),
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        ###########################################################################
        # Invoking in a test harness, where there's no __main__
        ###########################################################################
        # Explicit app properties, no metadata
        (
            EXPLICIT_FULL_APP_KWARGS,
            None,
            None,
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
        # Explicit app properties, but implied distribution name from app_id, no metadata
        (
            EXPLICIT_MIN_APP_KWARGS,
            None,
            None,
            "Explicit App",
            "org.beeware.explicit-app",
            "explicit-app",
        ),
        # No app properties, with metadata
        (
            dict(),
            APP_METADATA,
            None,
            "Test App",
            "org.beeware.test-app",
            "toga",
        ),
        # Explicit app properties, with metadata. Explicit values take precedence.
        (
            EXPLICIT_FULL_APP_KWARGS,
            APP_METADATA,
            None,
            "Explicit App",
            "org.beeware.explicit-app",
            "override-app",
        ),
    ],
)
def test_create(
    monkeypatch,
    event_loop,
    kwargs,
    metadata,
    main_module,
    expected_formal_name,
    expected_app_id,
    expected_app_name,
):
    """A simple app can be created."""
    # Monkeypatch the metadata retrieval function
    if metadata:
        metadata_mock = Mock(return_value=metadata)
    else:
        metadata_mock = Mock(
            side_effect=importlib.metadata.PackageNotFoundError(expected_app_name)
        )
    monkeypatch.setattr(importlib.metadata, "metadata", metadata_mock)

    # Monkeypatch the main module
    if main_module is None:
        try:
            monkeypatch.delitem(sys.modules, "__main__")
        except KeyError:
            pass
    else:
        monkeypatch.setitem(sys.modules, "__main__", main_module)

    app = toga.App(**kwargs)

    assert app.formal_name == expected_formal_name
    assert app.app_id == expected_app_id
    assert app.app_name == expected_app_name
    assert app.on_exit._raw is None

    metadata_mock.assert_called_once_with(expected_app_name)


@pytest.mark.parametrize(
    "kwargs, exc_type, message",
    [
        (dict(), RuntimeError, "Toga application must have a formal name"),
        (
            dict(formal_name="Something"),
            RuntimeError,
            "Toga application must have an app ID",
        ),
        (
            dict(windows=()),
            ValueError,
            "The `windows` constructor argument of toga.App has been removed",
        ),
    ],
)
def test_bad_app_creation(kwargs, exc_type, message):
    """Errors are raised."""
    with pytest.raises(exc_type, match=message):
        toga.App(**kwargs)


def test_app_metadata(monkeypatch, event_loop):
    """An app can load metadata from the .dist-info file."""
    monkeypatch.setattr(
        importlib.metadata,
        "metadata",
        Mock(
            return_value={
                "Formal-Name": "Metadata Name",
                "Name": "metadata",
                "App-ID": "org.beeware.metadata",
                "Author": "Jane Developer",
                "Version": "1.2.3",
                "Home-page": "https://example.com/test-app",
                "Summary": "A test app",
            }
        ),
    )

    # We can't use the app fixture here, because we need the metadata to be loaded as
    # part of app construction.
    app = toga.App(
        formal_name="Test App",
        app_id="org.example.test-app",
    )

    assert app.author == "Jane Developer"
    assert app.version == "1.2.3"
    assert app.home_page == "https://example.com/test-app"
    assert app.description == "A test app"
    assert app.is_bundled is False


def test_explicit_app_metadata(monkeypatch, event_loop):
    """App metadata can be provided explicitly, overriding module-level metadata."""
    monkeypatch.setattr(
        importlib.metadata,
        "metadata",
        Mock(
            return_value={
                "Formal-Name": "Metadata Name",
                "Name": "metadata",
                "App-ID": "org.beeware.metadata",
                "Author": "Alice Metadata",
                "Version": "2.3.4",
                "Home-page": "https://example.com/metadata",
                "Summary": "Metadata description of app",
            }
        ),
    )

    on_exit_handler = Mock()

    app = toga.App(
        formal_name="Test App",
        app_id="org.example.test-app",
        author="Jane Developer",
        version="1.2.3",
        home_page="https://example.com/test-app",
        description="A test app",
        on_exit=on_exit_handler,
    )

    assert app.author == "Jane Developer"
    assert app.version == "1.2.3"
    assert app.home_page == "https://example.com/test-app"
    assert app.description == "A test app"

    assert app.on_exit._raw == on_exit_handler


@pytest.mark.parametrize("construct", [True, False])
def test_icon_construction(construct, event_loop):
    """The app icon can be set during construction."""
    if construct:
        icon = toga.Icon("path/to/icon")
    else:
        icon = "path/to/icon"

    app = toga.App(
        formal_name="Test App",
        app_id="org.example.test",
        icon=icon,
    )
    assert isinstance(app.icon, toga.Icon)
    assert app.icon.path == Path("path/to/icon")


@pytest.mark.parametrize("construct", [True, False])
def test_icon(app, construct):
    """The app icon can be changed."""
    if construct:
        icon = toga.Icon("path/to/icon")
    else:
        icon = "path/to/icon"

    # Default icon matches distribution name
    assert isinstance(app.icon, toga.Icon)
    assert app.icon.path == Path("resources/test-app")
    # During initial setup, the icon isn't explicitly set.
    assert_action_not_performed(app, "set_icon")

    # Change icon
    app.icon = icon
    assert isinstance(app.icon, toga.Icon)
    assert app.icon.path == Path("path/to/icon")
    assert_action_performed_with(app, "set_icon", icon=toga.Icon("path/to/icon"))


def test_current_window(app):
    """The current window can be set and changed."""
    other_window = toga.Window()

    # There are two windows - the main window, plus "other"
    assert len(app.windows) == 2
    assert_action_performed_with(app, "set_main_window", window=app.main_window)

    # The initial current window is the main window
    assert app.current_window == app.main_window

    # Change the current window
    app.current_window = other_window
    assert app.current_window == other_window
    assert_action_performed_with(app, "set_current_window", window=other_window)


def test_no_current_window(app):
    """If there's no current window, current_window reflects this."""
    # If all the windows are deleted, and there's no main window (e.g., if it's a document app)
    # there might be no current window.
    app._main_window = None

    # The current window evaluates as None
    assert app.current_window is None


def test_full_screen(event_loop):
    """The app can be put into full screen mode."""
    window1 = toga.Window()
    window2 = toga.Window()
    app = toga.App(formal_name="Test App", app_id="org.example.test")

    assert not app.is_full_screen

    # If we're not full screen, exiting full screen is a no-op
    app.exit_full_screen()
    assert_action_not_performed(app, "exit_full_screen")

    # Enter full screen with 2 windows
    app.set_full_screen(window2, app.main_window)
    assert app.is_full_screen
    assert_action_performed_with(
        app, "enter_full_screen", windows=(window2, app.main_window)
    )

    # Change the screens that are full screen
    app.set_full_screen(app.main_window, window1)
    assert app.is_full_screen
    assert_action_performed_with(
        app, "enter_full_screen", windows=(app.main_window, window1)
    )

    # Exit full screen mode
    app.exit_full_screen()
    assert not app.is_full_screen
    assert_action_performed_with(
        app, "exit_full_screen", windows=(app.main_window, window1)
    )


def test_set_empty_full_screen_window_list(event_loop):
    """Setting the full screen window list to [] is an explicit exit."""
    app = toga.App(formal_name="Test App", app_id="org.example.test")
    window1 = toga.Window()
    window2 = toga.Window()

    assert not app.is_full_screen

    # Change the screens that are full screen
    app.set_full_screen(window1, window2)
    assert app.is_full_screen
    assert_action_performed_with(app, "enter_full_screen", windows=(window1, window2))

    # Exit full screen mode by setting no windows full screen
    app.set_full_screen()
    assert not app.is_full_screen
    assert_action_performed_with(app, "exit_full_screen", windows=(window1, window2))


def test_show_hide_cursor(app):
    """The app cursor can be shown and hidden."""
    app.hide_cursor()
    assert_action_performed(app, "hide_cursor")

    app.show_cursor()
    assert_action_performed(app, "show_cursor")


def test_startup_method(event_loop):
    """If an app provides a startup method, it will be invoked during startup."""
    startup = Mock()
    app = toga.App(
        formal_name="Test App",
        app_id="org.example.test",
        startup=startup,
    )

    startup.assert_called_once_with(app)


def test_startup_subclass(event_loop):
    """App can be subclassed."""

    class SubclassedApp(toga.App):
        def startup(self):
            self.main_window = toga.MainWindow()

    app = SubclassedApp(formal_name="Test App", app_id="org.example.test")

    # The main window will exist, and will have the app's formal name.
    assert app.main_window.title == "Test App"


def test_startup_subclass_no_main_window(event_loop):
    """If a subclassed app doesn't define a main window, an error is raised."""

    class SubclassedApp(toga.App):
        def startup(self):
            pass

    with pytest.raises(ValueError, match=r"Application does not have a main window."):
        SubclassedApp(formal_name="Test App", app_id="org.example.test")


def test_about(app):
    """The about dialog for the app can be shown."""
    app.about()
    assert_action_performed(app, "show_about_dialog")


def test_visit_homepage(monkeypatch, event_loop):
    """The app's homepage can be opened."""
    app = toga.App(
        formal_name="Test App",
        app_id="org.example.test",
        home_page="https://example.com/test-app",
    )
    open_webbrowser = Mock()
    monkeypatch.setattr(webbrowser, "open", open_webbrowser)

    # The app has no homepage by default, so visit is a no-op
    app.visit_homepage()

    open_webbrowser.assert_called_once_with("https://example.com/test-app")


def test_no_homepage(monkeypatch, app):
    """If the app doesn't have a home page, visit_homepage is a no-op."""
    open_webbrowser = Mock()
    monkeypatch.setattr(webbrowser, "open", open_webbrowser)

    # The app has no homepage by default, so visit is a no-op
    app.visit_homepage()

    open_webbrowser.assert_not_called()


def test_beep(app):
    """The machine can go Bing!"""
    app.beep()
    assert_action_performed(app, "beep")


def test_exit_direct(app):
    """An app can be exited directly."""
    on_exit_handler = Mock(return_value=True)
    app.on_exit = on_exit_handler

    # Exit the app directly
    app.exit()

    # App has been exited, but the exit handler has *not* been invoked.
    assert_action_performed(app, "exit")
    on_exit_handler.assert_not_called()


def test_exit_no_handler(app):
    """An app without an exit handler can be exited."""
    # Exit the app
    app._impl.simulate_exit()

    # Window has been exited, and is no longer in the app's list of windows.
    assert_action_performed(app, "exit")


def test_exit_successful_handler(app):
    """An app with a successful exit handler can be exited."""
    on_exit_handler = Mock(return_value=True)
    app.on_exit = on_exit_handler

    # Close the app
    app._impl.simulate_exit()

    # App has been exited
    assert_action_performed(app, "exit")
    on_exit_handler.assert_called_once_with(app)


def test_exit_rejected_handler(app):
    """An app can have a exit handler that rejects the exit."""
    on_exit_handler = Mock(return_value=False)
    app.on_exit = on_exit_handler

    # Close the window
    app._impl.simulate_exit()

    # App has been *not* exited
    assert_action_not_performed(app, "exit")
    on_exit_handler.assert_called_once_with(app)


def test_loop(app, event_loop):
    """The main thread's event loop can be accessed."""
    assert isinstance(app.loop, asyncio.AbstractEventLoop)
    assert app.loop is event_loop


def test_background_task(app):
    """A background task can be queued."""
    canary = Mock()

    async def background(app, **kwargs):
        canary()

    app.add_background_task(background)

    # Create an async task that we can use to start the event loop for a short time.
    async def waiter():
        await asyncio.sleep(0.1)

    app.loop.run_until_complete(waiter())

    # Once the loop has executed, the background task should have executed as well.
    canary.assert_called_once()


def test_deprecated_id(event_loop):
    """The deprecated `id` constructor argument is ignored, and the property of the same
    name is redirected to `app_id`"""
    id_warning = r"App.id is deprecated.* Use app_id instead"
    with pytest.warns(DeprecationWarning, match=id_warning):
        app = toga.App("Test App", "org.example.test", id="test_app_id")

    assert app.app_id == "org.example.test"
    with pytest.warns(DeprecationWarning, match=id_warning):
        assert app.id == "org.example.test"


def test_deprecated_name(event_loop):
    """The deprecated `name` property is redirected to `formal_name`"""
    name_warning = r"App.name is deprecated. Use formal_name instead"
    app = toga.App("Test App", "org.example.test")

    assert app.formal_name == "Test App"
    with pytest.warns(DeprecationWarning, match=name_warning):
        assert app.name == "Test App"



from unittest.mock import Mock

import pytest

import toga
from toga.types import Position
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def window():
    return toga.Window()


@pytest.fixture
def content():
    return toga.Box()


@pytest.fixture
def on_scroll_handler():
    return Mock()


@pytest.fixture
def scroll_container(content, on_scroll_handler):
    return toga.ScrollContainer(content=content, on_scroll=on_scroll_handler)


def test_widget_created():
    """A scroll container can be created with no arguments."""
    scroll_container = toga.ScrollContainer()
    assert scroll_container._impl.interface == scroll_container
    assert_action_performed(scroll_container, "create ScrollContainer")

    assert scroll_container.content is None
    assert scroll_container.vertical
    assert scroll_container.horizontal
    assert scroll_container.on_scroll._raw is None


def test_widget_created_with_values(content, on_scroll_handler):
    """A scroll container can be created with arguments."""
    scroll_container = toga.ScrollContainer(
        content=content,
        on_scroll=on_scroll_handler,
        vertical=False,
        horizontal=False,
    )
    assert scroll_container._impl.interface == scroll_container
    assert_action_performed(scroll_container, "create ScrollContainer")

    assert scroll_container.content == content
    assert not scroll_container.vertical
    assert not scroll_container.horizontal
    assert scroll_container.on_scroll._raw == on_scroll_handler

    # The content has been assigned to the widget
    assert_action_performed_with(
        scroll_container,
        "set content",
        widget=content._impl,
    )

    # The content has been refreshed
    assert_action_performed(content, "refresh")

    # The scroll handler hasn't been invoked
    on_scroll_handler.assert_not_called()


def test_assign_to_app(app, scroll_container, content):
    """If the widget is assigned to an app, the content is also assigned."""
    # Scroll container is initially unassigned
    assert scroll_container.app is None

    # Assign the scroll container to the app
    scroll_container.app = app

    # Scroll container is on the app
    assert scroll_container.app == app
    # Content is also on the app
    assert content.app == app


def test_assign_to_app_no_content(app):
    """If the widget is assigned to an app, and there is no content, there's no
    error."""
    scroll_container = toga.ScrollContainer()

    # Scroll container is initially unassigned
    assert scroll_container.app is None

    # Assign the scroll container to the app
    scroll_container.app = app

    # Scroll container is on the app
    assert scroll_container.app == app


def test_assign_to_window(window, scroll_container, content):
    """If the widget is assigned to a window, the content is also assigned."""
    # Scroll container is initially unassigned
    assert scroll_container.window is None

    # Assign the scroll container to the window
    scroll_container.window = window

    # Scroll container is on the window
    assert scroll_container.window == window
    # Content is also on the window
    assert content.window == window


def test_assign_to_window_no_content(window):
    """If the widget is assigned to an app, and there is no content, there's no
    error."""
    scroll_container = toga.ScrollContainer()

    # Scroll container is initially unassigned
    assert scroll_container.window is None

    # Assign the scroll container to the window
    scroll_container.window = window

    # Scroll container is on the window
    assert scroll_container.window == window


def test_disable_no_op(scroll_container):
    """ScrollContainer doesn't have a disabled state."""
    # Enabled by default
    assert scroll_container.enabled

    # Try to disable the widget
    scroll_container.enabled = False

    # Still enabled.
    assert scroll_container.enabled


def test_focus_noop(scroll_container):
    """Focus is a no-op."""

    scroll_container.focus()
    assert_action_not_performed(scroll_container, "focus")


def test_set_content(app, window, scroll_container, content):
    """The content of the scroll container can be changed."""
    # Assign the scroll container to an app and window
    scroll_container.app = app
    scroll_container.window = window

    # The content is also assigned
    assert content.app == app
    assert content.window == window

    # Reset the event log
    EventLog.reset()

    # Create new content, and assign it to the scroll container
    new_content = toga.Box()
    scroll_container.content = new_content

    # The content has been assigned to the widget
    assert_action_performed_with(
        scroll_container,
        "set content",
        widget=new_content._impl,
    )

    # The content has been refreshed
    assert_action_performed(new_content, "refresh")

    # The content has been assigned
    assert scroll_container.content == new_content

    # The new content is assigned to
    assert new_content.app == app
    assert new_content.window == window

    # The old content isn't
    assert content.app is None
    assert content.window is None


def test_clear_content(app, window, scroll_container, content):
    """The content of the scroll container can be cleared."""
    # Assign the scroll container to an app and window
    scroll_container.app = app
    scroll_container.window = window

    # The content is also assigned
    assert content.app == app
    assert content.window == window

    # Reset the event log
    EventLog.reset()

    # Clear the content
    scroll_container.content = None

    # The content has been assigned to the widget
    assert_action_performed_with(scroll_container, "set content", widget=None)

    # The content has been cleared
    assert scroll_container.content is None

    # The old content isn't assigned anymore
    assert content.app is None
    assert content.window is None


@pytest.mark.parametrize(
    "value, expected",
    [
        (True, True),
        (False, False),
        (42, True),
        (0, False),
        ("True", True),
        ("False", True),  # non-empty string is truthy
        ("", False),
        (object(), True),
    ],
)
def test_horizontal(scroll_container, on_scroll_handler, content, value, expected):
    """Horizontal scrolling can be enabled/disabled."""
    scroll_container.horizontal = value
    scroll_container.horizontal == expected

    if not expected:
        on_scroll_handler.assert_called_with(scroll_container)
    else:
        on_scroll_handler.assert_not_called()

    # Content is refreshed as a result of the change
    assert_action_performed(content, "refresh")


@pytest.mark.parametrize(
    "value, expected",
    [
        (True, True),
        (False, False),
        (42, True),
        (0, False),
        ("True", True),
        ("False", True),  # non-empty string is truthy
        ("", False),
        (object(), True),
    ],
)
def test_vertical(scroll_container, on_scroll_handler, content, value, expected):
    """Vertical scrolling can be enabled/disabled."""
    scroll_container.vertical = value
    scroll_container.vertical == expected

    if not expected:
        on_scroll_handler.assert_called_with(scroll_container)
    else:
        on_scroll_handler.assert_not_called()

    # Content is refreshed as a result of the change
    assert_action_performed(content, "refresh")


@pytest.mark.parametrize(
    "position, expected",
    [
        (10, 10),
        (-100, 0),  # Clipped to minimum value
        (1500, 1000),  # Clipped to maximum value
        ("10", 10),  # String, converted to int
        (10.1, 10),  # Float, converted to int
    ],
)
def test_horizontal_position(scroll_container, on_scroll_handler, position, expected):
    """The horizontal position can be set (clipped if necessary) and retrieved."""
    scroll_container.horizontal_position = position

    # scroll handler fired
    on_scroll_handler.assert_called_with(scroll_container)

    assert scroll_container.horizontal_position == expected
    assert scroll_container.max_horizontal_position == 1000


def test_disable_horizontal_scrolling(scroll_container, on_scroll_handler):
    """When disabling horizontal scrolling, horizontal position resets."""
    scroll_container.horizontal_position = 100
    on_scroll_handler.reset_mock()

    scroll_container.horizontal = False

    # scroll handler fired as a result of reset
    on_scroll_handler.assert_called_with(scroll_container)

    assert scroll_container.horizontal_position == 0
    assert scroll_container.max_horizontal_position == 0

    # Vertical position is unaffected by horizontal setting
    scroll_container.vertical_position = 100

    assert scroll_container.vertical_position == 100

    # scroll handler fired
    on_scroll_handler.assert_called_with(scroll_container)
    on_scroll_handler.reset_mock()


def test_horizontal_position_when_not_horizontal(scroll_container):
    """If horizontal scrolling isn't enabled, setting the horizontal position raises an
    error."""
    scroll_container.horizontal = False
    with pytest.raises(
        ValueError,
        match=r"Cannot set horizontal position when horizontal scrolling is not enabled.",
    ):
        scroll_container.horizontal_position = 37

    # horizontal coordinate is ignored when setting a full position
    scroll_container.position = (37, 42)

    assert scroll_container.horizontal_position == 0
    assert scroll_container.vertical_position == 42


@pytest.mark.parametrize(
    "position, expected",
    [
        (10, 10),
        (-100, 0),  # Clipped to minimum value
        (2500, 2000),  # Clipped to maximum value
        ("10", 10),  # String, converted to int
        (10.1, 10),  # Float, converted to int
    ],
)
def test_vertical_position(scroll_container, on_scroll_handler, position, expected):
    """The vertical position can be set (clipped if necessary) and retrieved."""
    scroll_container.vertical_position = position

    # scroll handler fired
    on_scroll_handler.assert_called_with(scroll_container)

    assert scroll_container.vertical_position == expected
    assert scroll_container.max_vertical_position == 2000


def test_disable_vertical_scrolling(scroll_container, on_scroll_handler):
    """When vertical scrolling is disabled, vertical position resets."""
    scroll_container.vertical_position = 100

    scroll_container.vertical = False

    # scroll handler fired as a result of reset
    on_scroll_handler.assert_called_with(scroll_container)
    on_scroll_handler.reset_mock()

    assert scroll_container.vertical_position == 0
    assert scroll_container.max_vertical_position == 0

    # Horizontal position is unaffected by vertical setting
    scroll_container.horizontal_position = 100

    assert scroll_container.horizontal_position == 100

    # scroll handler fired
    on_scroll_handler.assert_called_with(scroll_container)
    on_scroll_handler.reset_mock()


def test_set_vertical_position_when_not_vertical(scroll_container):
    """If vertical scrolling isn't enabled, setting the vertical position raises an
    error."""
    scroll_container.vertical = False
    with pytest.raises(
        ValueError,
        match=r"Cannot set vertical position when vertical scrolling is not enabled.",
    ):
        scroll_container.vertical_position = 42

    # vertical coordinate is ignored when setting a full position
    scroll_container.position = (37, 42)

    assert scroll_container.horizontal_position == 37
    assert scroll_container.vertical_position == 0


@pytest.mark.parametrize(
    "position, expected",
    [
        ((37, 42), Position(37, 42)),  # Cast to a Position type
        ((37, 42), (37, 42)),  # Backwards-compatible with tuple
        (Position(82, 82), Position(82, 82)),  # Accepts Position input
        ((-100, 42), (0, 42)),  # Clipped to minimum horizontal value
        ((37, -100), (37, 0)),  # Clipped to minimum vertical value
        ((-100, -100), (0, 0)),  # Clipped to minimum
        ((1500, 42), (1000, 42)),  # Clipped to maximum horizontal value
        ((37, 2500), (37, 2000)),  # Clipped to maximum vertical value
        ((1500, 2500), (1000, 2000)),  # Clipped to maximum
    ],
)
def test_position(scroll_container, on_scroll_handler, position, expected):
    """The scroll position can be set (clipped if necessary) and retrieved."""
    scroll_container.position = position

    assert scroll_container.position == expected

    # scroll handler fired
    on_scroll_handler.assert_called_with(scroll_container)



from __future__ import annotations

from collections.abc import Iterable

from .base import StyleT, Widget


class Box(Widget):
    _MIN_WIDTH = 0
    _MIN_HEIGHT = 0

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        children: Iterable[Widget] | None = None,
    ):
        """Create a new Box container widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param children: An optional list of children for to add to the Box.
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a Box
        self._impl = self.factory.Box(interface=self)

        # Children need to be added *after* the impl has been created.
        self._children: list[Widget] = []
        if children is not None:
            self.add(*children)

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        Box widgets cannot be disabled; this property will always return True; any
        attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: bool) -> None:
        pass

    def focus(self) -> None:
        """No-op; Box cannot accept input focus."""
        pass



from unittest.mock import Mock

import pytest

import toga


@pytest.fixture
def on_resize_handler():
    return Mock()


@pytest.fixture
def on_press_handler():
    return Mock()


@pytest.fixture
def on_drag_handler():
    return Mock()


@pytest.fixture
def on_release_handler():
    return Mock()


@pytest.fixture
def on_activate_handler():
    return Mock()


@pytest.fixture
def on_alt_press_handler():
    return Mock()


@pytest.fixture
def on_alt_drag_handler():
    return Mock()


@pytest.fixture
def on_alt_release_handler():
    return Mock()


@pytest.fixture
def widget(
    on_resize_handler,
    on_press_handler,
    on_activate_handler,
    on_release_handler,
    on_drag_handler,
    on_alt_press_handler,
    on_alt_release_handler,
    on_alt_drag_handler,
):
    return toga.Canvas(
        on_resize=on_resize_handler,
        on_press=on_press_handler,
        on_activate=on_activate_handler,
        on_release=on_release_handler,
        on_drag=on_drag_handler,
        on_alt_press=on_alt_press_handler,
        on_alt_release=on_alt_release_handler,
        on_alt_drag=on_alt_drag_handler,
    )
